"""PDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

This file contains term-related object_action functions for processing Egeria Markdown


"""

import json
import os
import sys
from typing import Optional

from loguru import logger
from rich import print
from rich.console import Console
from rich.markdown import Markdown

from md_processing.md_processing_utils.common_md_proc_utils import (parse_upsert_command, parse_view_command )
from md_processing.md_processing_utils.common_md_utils import (
    update_element_dictionary,
    set_find_body,
    set_create_body,
    set_element_prop_body,
    set_update_body,
    set_rel_request_body,
    set_update_status_body,
    set_rel_prop_body,
    set_data_field_body,
)
# Developer note: Request bodies follow a two-layer pattern:
# - Outer: action wrapper (NewElementRequestBody/UpdateElementRequestBody/...)
# - Inner: element-specific Properties dict placed under the outer["properties"] key
# Always build inner bodies via helpers (set_element_prop_body, set_data_field_body, etc.) and assign to the outer
from md_processing.md_processing_utils.extraction_utils import (extract_command_plus, update_a_command)
from md_processing.md_processing_utils.md_processing_constants import (load_commands, ERROR)
from pyegeria import DEBUG_LEVEL, body_slimmer
from pyegeria.egeria_tech_client import EgeriaTech
from pyegeria.utils import make_format_set_name_from_type

GERIA_METADATA_STORE = os.environ.get("EGERIA_METADATA_STORE", "active-metadata-store")
EGERIA_KAFKA_ENDPOINT = os.environ.get("KAFKA_ENDPOINT", "localhost:9092")
EGERIA_PLATFORM_URL = os.environ.get("EGERIA_PLATFORM_URL", "https://localhost:9443")
EGERIA_VIEW_SERVER = os.environ.get("EGERIA_VIEW_SERVER", "view-server")
EGERIA_VIEW_SERVER_URL = os.environ.get("EGERIA_VIEW_SERVER_URL", "https://localhost:9443")
EGERIA_INTEGRATION_DAEMON = os.environ.get("EGERIA_INTEGRATION_DAEMON", "integration-daemon")
EGERIA_INTEGRATION_DAEMON_URL = os.environ.get("EGERIA_INTEGRATION_DAEMON_URL", "https://localhost:9443")
EGERIA_ADMIN_USER = os.environ.get("ADMIN_USER", "garygeeke")
EGERIA_ADMIN_PASSWORD = os.environ.get("ADMIN_PASSWORD", "secret")
EGERIA_USER = os.environ.get("EGERIA_USER", "erinoverview")
EGERIA_USER_PASSWORD = os.environ.get("EGERIA_USER_PASSWORD", "secret")
EGERIA_WIDTH = os.environ.get("EGERIA_WIDTH", 220)
EGERIA_JUPYTER = os.environ.get("EGERIA_JUPYTER", False)
EGERIA_HOME_GLOSSARY_GUID = os.environ.get("EGERIA_HOME_GLOSSARY_GUID", None)
EGERIA_GLOSSARY_PATH = os.environ.get("EGERIA_GLOSSARY_PATH", None)
EGERIA_ROOT_PATH = os.environ.get("EGERIA_ROOT_PATH", "../../")
EGERIA_INBOX_PATH = os.environ.get("EGERIA_INBOX_PATH", "md_processing/dr_egeria_inbox")
EGERIA_OUTBOX_PATH = os.environ.get("EGERIA_OUTBOX_PATH", "md_processing/dr_egeria_outbox")

load_commands('commands.json')
debug_level = DEBUG_LEVEL

console = Console(width=int(200))


#
# Helper functions for the data designer commands
#
@logger.catch
def add_member_to_data_collections(egeria_client: EgeriaTech, collection_list: list, display_name: str,
                                   guid: str) -> None:
    """
    Add member to data dictionaries and data specifications.
    """
    body = {
        "class": "NewRelationshipRequestBody", "properties": {
            "class": "CollectionMembershipProperties", "membershipRationale": "User Specified",
            "notes": "Added by Dr.Egeria"
            }
        }
    try:
        if collection_list is not None:
            for collection in collection_list:
                egeria_client.add_to_collection(collection, guid, body)
                msg = f"Added `{display_name}` member to `{collection}`"
                logger.info(msg)
        else:
            logger.info("There were no data collections to add.")
        return

    except Exception as e:
        console.print_exception()


@logger.catch
def remove_member_from_data_collections(egeria_client: EgeriaTech, collection_list: list, display_name: str,
                                        guid: str) -> None:
    try:
        for collection in collection_list:
            egeria_client.remove_from_collection(collection, guid)
            msg = f"Removed `{display_name}` member from `{collection}`"
            logger.info(msg)
        return

    except Exception as e:
        console.print_exception()


@logger.catch
def update_data_collection_memberships(egeria_client: EgeriaTech, entity_type: str, guid_list: list,
                                       collection_class: str, guid: str, display_name: str,
                                       replace_all_props: bool = True) -> None:
    """ update the collection membership of the element

        If replace_all_props is set to True, all existing memberships are removed and new memberships are added.
        If replace_all_props is set to False, only the new memberships are added.
    """

    if replace_all_props:
        match entity_type:
            case "Data Specification":
                get_command = egeria_client.get_collection_by_guid
            case "Data Structure":
                get_command = egeria_client.get_data_structure_by_guid
            case "Data Field":
                get_command = egeria_client.get_data_field_by_guid
            case "Data Class":
                get_command = egeria_client.get_data_class_by_guid

        coll_list = egeria_client.get_data_memberships(get_command, guid)
        if coll_list is None:
            logger.warning("Unexpected -> the collection list was None - assigning empty dict")
            coll_list = {}
        # compare the existing collections to desired collections
        if collection_class == "DataDictionary":
            as_is = set(coll_list.get("DictList", {}))
        elif collection_class == "DataSpec":
            as_is = set(coll_list.get("SpecList", {}))

        dict_set = set(coll_list.get("DictList", {}))
        spec_set = set(coll_list.get("SpecList", {}))
        to_be_set = set(guid_list) if guid_list is not None else set()
        logger.debug(f"as_is: {as_is}")
        logger.debug(f"to_be_set: {to_be_set}")

        # Remove membership for collections that are in the as-is but not in the to-be
        to_remove = as_is - to_be_set
        logger.debug(f"to_remove: {to_remove}")
        if len(to_remove) > 0:
            remove_member_from_data_collections(egeria_client, to_remove, display_name, guid)

        # add membership for collections that are in the to-be but are not in the as-is
        to_add = to_be_set - as_is
        logger.debug(f"to_add: {to_add}")
        if len(to_add) > 0:
            add_member_to_data_collections(egeria_client, to_add, display_name, guid)
    else:
        add_member_to_data_collections(egeria_client, guid_list, display_name, guid)


# @logger.catch


@logger.catch
def add_field_to_data_structures(egeria_client: EgeriaTech, display_name: str, struct_list: list, guid) -> None:
    """
        Add data field to data structures.
        """

    try:
        for structure_guid in struct_list:
            egeria_client.link_member_data_field(structure_guid, guid, None)
            msg = f"Added `{display_name}` to structure `{structure_guid}`"
            logger.info(msg)
        return

    except Exception as e:
        console.print_exception()


@logger.catch
def remove_field_from_data_structures(egeria_client: EgeriaTech, display_name: str, struct_list: list,
                                      guid: str) -> None:
    """Remove a data field from a list of data structures."""
    try:
        for structure_guid in struct_list:
            egeria_client.detach_member_data_field(structure_guid, guid, None)
            msg = f"Removed `{display_name}` from structure `{structure_guid}`"
            logger.info(msg)
        return

    except Exception as e:
        console.print_exception()


@logger.catch
def sync_data_field_rel_elements(egeria_client: EgeriaTech, structure_list: list, parent_field_list: list, terms: list,
                                 data_class_guid: str, guid: str, display_name: str,
                                 replace_all_props: bool = True) -> None:
    """Sync a field's related elements.

    TODO: Need to add data class support when ready and may need to revisit bodies.

    """
    if terms:
        terms = [terms]

    if replace_all_props:
        rel_el_list = egeria_client.get_data_field_rel_elements(guid)
        # should I throw an exception if empty?
        if rel_el_list is None:
            logger.warning("Unexpected -> the list was None - assigning empty list")
            rel_el_list = {}

        as_is_data_structs = set(rel_el_list.get("data_structure_guids", []))
        as_is_parent_fields = set(rel_el_list.get("parent_guids", []))
        as_is_assigned_meanings = set(rel_el_list.get("assigned_meanings_guids", []))
        as_is_data_classes = set(rel_el_list.get("data_class_guids", []))

        to_be_data_structs = set(structure_list) if structure_list is not None else set()
        to_be_parent_fields = set(parent_field_list) if parent_field_list is not None else set()
        to_be_assigned_meanings = set(terms) if terms is not None else set()
        to_be_data_classes = set([data_class_guid]) if data_class_guid is not None else set()

        logger.trace(f"as_is_data_structs: {list(as_is_data_structs)} to_be_data_struct: {list(to_be_data_structs)}")
        logger.trace(
            f"as_is_parent_fields: {list(as_is_parent_fields)} to_be_parent_fields: {list(to_be_parent_fields)}")
        logger.trace(f"as_is_assigned_meanings: {list(as_is_assigned_meanings)} to_be_assigned_meanings: "
                     f"{list(to_be_assigned_meanings)}")
        logger.trace(f"as_is_data_classes: {list(as_is_data_classes)} to_be_assigned_data_classes: "
                     f"{list(to_be_data_classes)}")

        data_struct_to_remove = as_is_data_structs - to_be_data_structs
        logger.trace(f"data_struct_to_remove: {list(data_struct_to_remove)}")
        if len(data_struct_to_remove) > 0:
            for ds in data_struct_to_remove:
                egeria_client.detach_member_data_field(ds, guid, None)
                msg = f"Removed `{display_name}` from structure `{ds}`"
                logger.trace(msg)
        data_struct_to_add = to_be_data_structs - as_is_data_structs
        logger.trace(f"data_struct_to_add: {list(data_struct_to_add)}")
        if len(data_struct_to_add) > 0:
            for ds in data_struct_to_add:
                egeria_client.link_member_data_field(ds, guid, None)
                msg = f"Added `{display_name}` to structure `{ds}`"
                logger.trace(msg)

        parent_field_to_remove = to_be_parent_fields - as_is_parent_fields
        logger.trace(f"parent_field_to_remove: {list(parent_field_to_remove)}")
        if len(parent_field_to_remove) > 0:
            for field in parent_field_to_remove:
                egeria_client.detach_nested_data_field(field, guid, None)
                msg = f"Removed `{display_name}` from field `{field}`"
                logger.trace(msg)
        parent_field_to_add = to_be_parent_fields - as_is_parent_fields
        logger.trace(f"parent_field_to_add: {list(parent_field_to_add)}")
        if len(parent_field_to_add) > 0:
            for field in parent_field_to_add:
                egeria_client.link_nested_data_field(field, guid, None)
                msg = f"Added `{display_name}` to field `{field}`"
                logger.trace(msg)

        terms_to_remove = as_is_assigned_meanings - to_be_assigned_meanings
        logger.trace(f"terms_to_remove: {list(terms_to_remove)}")
        if terms:
            for term in terms_to_remove:
                egeria_client.detach_semantic_definition(guid, term, None)
                msg = f"Removed `{term}` from `{display_name}`"
                logger.trace(msg)
        terms_to_add = to_be_assigned_meanings - as_is_assigned_meanings
        logger.trace(f"terms_to_add: {list(terms_to_add)}")
        if len(terms_to_add) > 0:
            for term in terms_to_add:
                egeria_client.link_semantic_definition(guid, term, None)
                msg = f"Added `{term}` to`{display_name}`"
                logger.trace(msg)

        classes_to_remove = as_is_data_classes - to_be_data_classes
        logger.trace(f"classes_to_remove: {list(classes_to_remove)}")
        if len(terms_to_remove) > 0:
            for dc in classes_to_remove:
                body = {
                    "class": "MetadataSourceRequestBody", "forLineage": False, "forDuplicateProcessing": False
                    }
                egeria_client.detach_data_class_definition(guid, dc, body)
                msg = f"Removed `{dc}` from `{display_name}`"
                logger.trace(msg)
        classes_to_add = to_be_data_classes - as_is_data_classes
        logger.trace(f"classes_to_add: {list(classes_to_add)}")
        if len(terms_to_add) > 0:
            for dc in classes_to_add:
                body = {
                    "class": "RelationshipRequestBody", "forLineage": False, "forDuplicateProcessing": False
                    }
                egeria_client.link_data_class_definition(guid, dc, body)
                msg = f"Added `{dc}` to`{display_name}`"
                logger.trace(msg)


    else:  # merge - add field to related elements
        if structure_list:
            add_field_to_data_structures(egeria_client, display_name, structure_list, guid)
            msg = f"Added `{display_name}` to `{structure_list}`"
            logger.trace(msg)

        if parent_field_list:
            for field in parent_field_list:
                egeria_client.link_nested_data_field(field, guid, None)
                msg = f"Added `{display_name}` to `{field}`"
                logger.trace(msg)
        if terms:
            for term in terms:
                egeria_client.link_semantic_definition(guid, term, None)
                msg = f"Added `{term}` to `{display_name}`"
                logger.trace(msg)

        if data_class_guid:
            egeria_client.link_data_class_definition(guid, data_class_guid)
            msg = f"Added `{data_class_guid}` to `{display_name}`"
            logger.trace(msg)


@logger.catch
def sync_data_class_rel_elements(egeria_client: EgeriaTech, containing_data_class_guids: list, terms: list,
                                 specializes_data_classes: list, guid: str, display_name: str,
                                 replace_all_props: bool = True) -> None:
    """Sync a data class' related elements.

    """
    if terms:
        terms = [terms]

    if replace_all_props:
        rel_el_list = egeria_client.get_data_class_rel_elements(guid)
        if rel_el_list is None:
            logger.warning("Unexpected -> the list was None - assigning empty list")
            rel_el_list = {}
        if terms:
            terms = [terms]

        as_is_nested_classes = set(rel_el_list.get("nested_data_class_guids", []))
        as_is_assigned_meanings = set(rel_el_list.get("assigned_meanings_guids", []))
        as_is_specialized_classes = set(rel_el_list.get("specialized_data_class_guids", []))

        to_be_nested_classes = set(containing_data_class_guids) if containing_data_class_guids is not None else set()
        to_be_assigned_meanings = set(terms) if terms is not None else set()
        to_be_specialized_classes = set([specializes_data_classes]) if specializes_data_classes is not None else set()

        logger.trace(
            f"as_is_nested_classes: {list(as_is_nested_classes)} to_be_nested_classes: {list(to_be_nested_classes)}")
        logger.trace(f"as_is_assigned_meanings: {list(as_is_assigned_meanings)} to_be_assigned_meanings: "
                     f"{list(to_be_assigned_meanings)}")
        logger.trace(f"as_is_specialized_classes: {list(as_is_specialized_classes)} to_be_specizialized_data_classes: "
                     f"{list(to_be_specialized_classes)}")

        nested_classes_to_remove = to_be_nested_classes - as_is_nested_classes
        logger.trace(f"nested_classes_to_remove: {list(nested_classes_to_remove)}")
        if len(nested_classes_to_remove) > 0:
            for field in nested_classes_to_remove:
                egeria_client.detach_nested_data_class(field, guid, None)
                msg = f"Removed `{display_name}` from field `{field}`"
                logger.trace(msg)
        nested_classes_to_add = to_be_nested_classes - as_is_nested_classes
        logger.trace(f"nested_classes_to_add: {list(nested_classes_to_add)}")
        if len(nested_classes_to_add) > 0:
            for field in nested_classes_to_add:
                egeria_client.link_nested_data_class(field, guid, None)
                msg = f"Added `{display_name}` to field `{field}`"
                logger.trace(msg)

        terms_to_remove = as_is_assigned_meanings - to_be_assigned_meanings
        logger.trace(f"terms_to_remove: {list(terms_to_remove)}")
        if len(terms_to_remove) > 0:
            for term in terms_to_remove:
                egeria_client.detach_semantic_definition(guid, term, None)
                msg = f"Removed `{term}` from `{display_name}`"
                logger.trace(msg)
        terms_to_add = to_be_assigned_meanings - as_is_assigned_meanings
        logger.trace(f"terms_to_add: {list(terms_to_add)}")
        if len(terms_to_add) > 0:
            for term in terms_to_add:
                egeria_client.link_semantic_definition(guid, term, None)
                msg = f"Added `{term}` to`{display_name}`"
                logger.trace(msg)

        specialized_classes_to_remove = as_is_specialized_classes - to_be_specialized_classes
        logger.trace(f"classes_to_remove: {list(specialized_classes_to_remove)}")
        if len(terms_to_remove) > 0:
            for dc in specialized_classes_to_remove:
                body = {
                    "class": "MetadataSourceRequestBody", "forLineage": False, "forDuplicateProcessing": False
                    }
                egeria_client.detach_specialist_data_class(guid, dc, body)
                msg = f"Removed `{dc}` from `{display_name}`"
                logger.trace(msg)
        specialized_classes_to_add = to_be_specialized_classes - as_is_specialized_classes
        logger.trace(f"classes_to_add: {list(specialized_classes_to_add)}")
        if len(specialized_classes_to_add) > 0:
            for dc in specialized_classes_to_add:
                body = {
                    "class": "RelationshipRequestBody", "forLineage": False, "forDuplicateProcessing": False
                    }
                egeria_client.link_specialist_data_class(guid, dc, body)
                msg = f"Added `{dc}` to`{display_name}`"
                logger.trace(msg)


    else:  # merge - add field to related elements
        if containing_data_class_guids:
            for field in containing_data_class_guids:
                egeria_client.link_nested_data_class(field, guid, None)
                msg = f"Added `{display_name}` to `{field}`"
                logger.trace(msg)

        if terms:
            for term in terms:
                egeria_client.link_semantic_definition(guid, term, None)
                msg = f"Added `{term}` to `{display_name}`"
                logger.trace(msg)
        if specializes_data_classes:
            for el in specializes_data_classes:
                egeria_client.link_specialist_data_class(guid, el)
            msg = f"Linked `{el}` to `{display_name}`"
            logger.trace(msg)


@logger.catch
def process_data_spec_upsert_command(egeria_client: EgeriaTech, txt: str, directive: str = "display") -> Optional[str]:
    """
    Processes a data specification create or update object_action by extracting key attributes such as
    spec name, parent_guid, parent_relationship_type, parent_at_end_1, category

    :param txt: A string representing the input cell to be processed for
        extracting glossary-related attributes.
    :param directive: an optional string indicating the directive to be used - display, validate or execute
    :return: A string summarizing the outcome of the processing.
    """

    command, object_type, object_action = extract_command_plus(txt)
    print(Markdown(f"# {command}\n"))

    parsed_output = parse_upsert_command(egeria_client, object_type, object_action, txt, directive)

    valid = parsed_output['valid']
    exists = parsed_output['exists']

    qualified_name = parsed_output.get('qualified_name', None)
    guid = parsed_output.get('guid', None)

    logger.debug(json.dumps(parsed_output, indent=4))

    attributes = parsed_output['attributes']
    display_name = attributes.get('Display Name', {}).get('value', "None Found")
    status = attributes.get('Status', {}).get('value', None)
    print(Markdown(parsed_output['display']))

    in_data_spec_list = attributes.get('In Data Specification', {}).get('value', None)
    in_data_spec_valid = attributes.get('In Data Specification', {}).get('valid', None)
    in_data_spec_exists = attributes.get('In Data Specification', {}).get('exists', None)
    output_set = make_format_set_name_from_type(object_type)
    object_type = "DataSpec"
    print(Markdown(parsed_output['display']))

    if directive == "display":

        return None
    elif directive == "validate":
        if valid:
            print(Markdown(f"==> Validation of {command} completed successfully!\n"))
        else:
            msg = f"Validation failed for object_action `{command}`\n"
        return valid

    elif directive == "process":

        try:
            if object_action == "Update":
                if not exists:
                    msg = (f" Element `{display_name}` does not exist! Updating result document with Create "
                           f"object_action\n")
                    logger.error(msg)
                    return update_a_command(txt, object_action, object_type, qualified_name, guid)
                elif not valid:
                    return None
                else:
                    print(Markdown(
                        f"==> Validation of {command} completed successfully! Proceeding to apply the changes.\n"))

                body = set_update_body(object_type, attributes)
                body['properties'] = set_element_prop_body(object_type, qualified_name, attributes)

                egeria_client.update_collection(guid, body)
                if status:
                    egeria_client.update_collection_status(guid, status)

                logger.success(f"Updated  {object_type} `{display_name}` with GUID {guid}\n\n___")
                update_element_dictionary(qualified_name, {
                    'guid': guid, 'display_name': display_name
                })
                return egeria_client.get_collection_by_guid(guid, element_type='Data-Specification-DrE',
                                                            output_format='MD', report_spec=output_set)

            elif object_action == "Create":
                if valid is False and exists:
                    msg = (f"  Data Specification `{display_name}` already exists and result document updated changing "
                           f"`Create` to `Update` in processed output\n\n___")
                    logger.error(msg)
                    return update_a_command(txt, object_action, object_type, qualified_name, guid)
                elif valid is False and in_data_spec_valid is False:
                    msg = (f" Invalid data specification(s) `{in_data_spec_list}` "
                           f" invalid data? - Correct and try again\n\n___")
                    logger.error(msg)
                    return None
                else:
                    body = set_create_body(object_type, attributes)
                    body["properties"] = set_element_prop_body(object_type, qualified_name, attributes)
                    parent_guid = body.get('parentGuid', None)
                    if parent_guid:
                        body['parentRelationshipTypeName'] = "CollectionMembership"
                        body['parentAtEnd1'] = True

                    guid = egeria_client.create_collection(body=body)
                    if guid:
                        update_element_dictionary(qualified_name, {
                            'guid': guid, 'display_name': display_name
                        })
                        msg = f"Created Element `{display_name}` with GUID {guid}\n\n___"
                        logger.success(msg)
                        return egeria_client.get_collection_by_guid(guid, object_type, output_format='MD', report_spec=output_set)
                    else:
                        msg = f"Failed to create element `{display_name}` with GUID {guid}\n\n___"
                        logger.error(msg)
                        return None


        except Exception as e:
            logger.error(f"Error performing {command}: {e}")
            return None
    else:
        return None


@logger.catch
def process_data_dict_upsert_command(egeria_client: EgeriaTech, txt: str, directive: str = "display") -> Optional[str]:
    """
    Processes a data dictionary create or update object_action by extracting key attributes such as
    spec name, parent_guid, parent_relationship_type, parent_at_end_1, category

    :param txt: A string representing the input cell to be processed for
        extracting glossary-related attributes.
    :param directive: an optional string indicating the directive to be used - display, validate or execute
    :return: A string summarizing the outcome of the processing.
    """


    command, object_type, object_action = extract_command_plus(txt)
    print(Markdown(f"# {command}\n"))

    parsed_output = parse_upsert_command(egeria_client, object_type, object_action, txt, directive)

    valid = parsed_output['valid']
    exists = parsed_output['exists']

    qualified_name = parsed_output.get('qualified_name', None)
    guid = parsed_output.get('guid', None)

    print(Markdown(parsed_output['display']))

    logger.debug(json.dumps(parsed_output, indent=4))

    attributes = parsed_output['attributes']
    display_name = attributes.get('Display Name', {}).get('value', "None Found")
    status = attributes.get('Status', {}).get('value', None)
    output_set = make_format_set_name_from_type(object_type)

    if directive == "display":

        return None
    elif directive == "validate":
        if valid:
            print(Markdown(f"==> Validation of {command} completed successfully!\n"))
        else:
            msg = f"Validation failed for object_action `{command}`\n"
        return valid

    elif directive == "process":

        try:
            if object_action == "Update":
                if not exists:
                    msg = (f" Element `{display_name}` does not exist! Updating result document with Create "
                           f"object_action\n")
                    logger.error(msg)
                    return update_a_command(txt, object_action, object_type, qualified_name, guid)
                elif not valid:
                    return None
                else:
                    print(Markdown(
                        f"==> Validation of {command} completed successfully! Proceeding to apply the changes.\n"))

                body = set_update_body(object_type, attributes)
                body['properties'] = set_element_prop_body("Data Spec", qualified_name, attributes)

                egeria_client.update_collection(guid, body)
                if status:
                    egeria_client.update_collection_status(guid, status)

                logger.success(f"Updated  {object_type} `{display_name}` with GUID {guid}\n\n___")
                update_element_dictionary(qualified_name, {
                    'guid': guid, 'display_name': display_name
                })
                return egeria_client.get_collection_by_guid(guid, element_type='Data Specification',
                                                            output_format='MD', report_spec=output_set)

            elif object_action == "Create":
                if valid is False and exists:
                    msg = (f"  Data Specification `{display_name}` already exists and result document updated changing "
                           f"`Create` to `Update` in processed output\n\n___")
                    logger.error(msg)
                    return update_a_command(txt, object_action, object_type, qualified_name, guid)
                elif valid is False :
                    msg = f" invalid data? - Correct and try again\n\n___"
                    logger.error(msg)
                    return None
                else:
                    body = set_create_body(object_type, attributes)
                    body["properties"] = set_element_prop_body(object_type, qualified_name, attributes)
                    parent_guid = body.get('parentGuid', None)
                    if parent_guid:
                        body['parentRelationshipTypeName'] = "CollectionMembership"
                        body['parentAtEnd1'] = True

                    guid = egeria_client.create_collection(body=body)
                    if guid:
                        update_element_dictionary(qualified_name, {
                            'guid': guid, 'display_name': display_name
                        })
                        msg = f"Created Element `{display_name}` with GUID {guid}\n\n___"
                        logger.success(msg)
                        return egeria_client.get_collection_by_guid(guid, object_type, output_format='MD', report_spec=output_set)
                    else:
                        msg = f"Failed to create element `{display_name}` with GUID {guid}\n\n___"
                        logger.error(msg)
                        return None


        except Exception as e:
            logger.error(f"Error performing {command}: {e}")
            return None
    else:
        return None


@logger.catch
def process_data_structure_upsert_command(egeria_client: EgeriaTech, txt: str,
                                          directive: str = "display") -> Optional[str]:
    """
    Processes a data structure create or update object_action by extracting key attributes such as
    spec name, parent_guid, parent_relationship_type, parent_at_end_1, category

    :param txt: A string representing the input cell to be processed for
        extracting glossary-related attributes.
    :param directive: an optional string indicating the directive to be used - display, validate or execute
    :return: A string summarizing the outcome of the processing.
    """


    command, object_type, object_action = extract_command_plus(txt)
    print(Markdown(f"# {command}\n"))

    parsed_output = parse_upsert_command(egeria_client, object_type, object_action, txt, directive)

    valid = parsed_output['valid']
    exists = parsed_output['exists']

    qualified_name = parsed_output.get('qualified_name', None)
    guid = parsed_output.get('guid', None)
    output_set = make_format_set_name_from_type(object_type)

    print(Markdown(parsed_output['display']))

    if directive == "display":
        return None

    elif directive == "validate":
        if valid:
            print(Markdown(f"==> Validation of {command} completed successfully!\n"))
        else:
            msg = f"Validation failed for object_action `{command}`\n"
        return valid

    elif directive == "process":
        logger.debug(json.dumps(parsed_output, indent=4))
        attributes = parsed_output['attributes']
        display_name = attributes.get('Display Name', {}).get('value', None)
        qualified_name = attributes.get('Qualified Name',{}).get('value', None)
        if qualified_name is None:
            qualified_name = attributes.get('Display Name',{}).get('qualified_name', None)
            if qualified_name is None:
                logger.error(f"Element `{display_name}` does not have a valid specification? No qualified name..Review..\n\n___ ")
                return None
        prop_body = set_element_prop_body(object_type, qualified_name, attributes)
        prop_body['namespace'] = attributes.get('Namespace', {}).get('value', None)
        display_name = attributes.get('Display Name',{}).get('value', None)

        data_spec_name_list = attributes.get("In Data Specification", {}).get("name_list", "")
        data_spec_value = attributes.get("In Data Specification", {}).get("value", None)
        data_spec_guid_list = attributes.get("In Data Specification", {}).get("guid_list", None)

        in_data_dictionary = attributes.get('In Data Dictionary', {}).get('dict_list', None)
        data_dict_name_list = attributes.get('In Data Dictionary', {}).get('name_list', "")
        data_dict_value_list = attributes.get('In Data Dictionary', {}).get('value', None)
        data_dict_guid_list = attributes.get("In Data Dictionary", {}).get("guid_list", None)
        merge_update = attributes.get('Merge Update', {}).get('value', True)

        try:
            if object_action == "Update":
                if not exists:
                    logger.error(f"Element `{qualified_name}` does not exist! Updating result document with Create "
                                 f"object_action\n\n___")
                    return update_a_command(txt, object_action, object_type, qualified_name, guid)
                elif not valid:
                    logger.error(f"Element `{qualified_name}` does not have a valid specification? Review..\n\n___ ")
                    return None
                else:
                    update_body = set_update_body(object_type,attributes)
                    update_body['properties'] = prop_body
                    egeria_client.update_data_structure(guid, update_body)
                    update_element_dictionary(qualified_name, {
                        'guid': guid, 'display_name': display_name
                        })
                    logger.success(f"Updated  {object_type} `{display_name}` with GUID {guid}\n\n___")

                    core_props = egeria_client.get_data_structure_by_guid(guid, output_format='MD', report_spec=output_set)

                    update_element_dictionary(qualified_name, {
                        'guid': guid, 'display_name': display_name
                    })

                    update_data_collection_memberships(egeria_client, object_type, data_spec_guid_list, "DataSpec", guid,
                                                       display_name, merge_update)
                    update_data_collection_memberships(egeria_client, object_type, data_dict_guid_list, "DataSpec",
                                                       guid,display_name, merge_update)
                    core_props += f"## In Data Dictionary\n\n{data_dict_name_list}\n\n"
                    core_props += f"## In Data Specification\n\n{data_spec_name_list}\n\n"
                    logger.success(f"Updated  {object_type} `{display_name}` with GUID {guid}\n\n___")
                    return core_props


            if not valid:
                if exists and object_action == "Create":
                    msg = (f"Create failed because Element `{display_name}` exists - changing `Create` to `Update` in "
                           f"processed output \n\n___")
                    logger.error(msg)
                    return update_a_command(txt, object_action, object_type, qualified_name, guid)
                else:
                    return None


            elif object_action == "Create":
                if exists:
                    logger.warning(f"\nTerm `{display_name}` already exists and result document updated\n\n___")
                    return update_a_command(txt, object_action, object_type, qualified_name, guid)
                else:
                    body = set_create_body(object_type, attributes)
                    body['properties'] = prop_body
                    guid = egeria_client.create_data_structure(body_slimmer(body))
                    if guid:
                        update_element_dictionary(qualified_name, {
                            'guid': guid, 'display_name': display_name
                            })

                        core_props = egeria_client.get_data_structure_by_guid(guid, output_format='MD', report_spec=output_set)

                        if in_data_dictionary:
                            logger.info(f"Will add to data dictionary(s) `{in_data_dictionary}`")
                            add_member_to_data_collections(egeria_client, data_dict_guid_list,
                                                                    display_name,guid)
                            core_props += f"## In Data Dictionary\n\n{data_dict_name_list}\n\n"

                        if data_spec_guid_list:
                            add_member_to_data_collections(egeria_client, data_spec_guid_list,
                                                                    display_name, guid)
                            core_props += f"## In Data Specifications\n\n`{data_spec_name_list}`\n\n"

                        logger.info(f"Created Element `{display_name}` with GUID {guid}\n\n___")

                        return core_props
                    else:
                        logger.error(f"Failed to create Data Structure `{display_name}`\n\n___")
                        return None

        except Exception as e:
            logger.error(f"Error performing {object_action}: {e}\n\n___")
            return None
        else:
            return None


@logger.catch
def process_data_field_upsert_command(egeria_client: EgeriaTech, txt: str, directive: str = "display") -> Optional[str]:
    """
    Processes a data field create or update object_action by extracting key attributes such as
    spec name, parent_guid, parent_relationship_type, parent_at_end_1, category

    :param txt: A string representing the input cell to be processed for
        extracting glossary-related attributes.
    :param directive: an optional string indicating the directive to be used - display, validate or execute
    :return: A string summarizing the outcome of the processing.
    """
    from md_processing.md_processing_utils.common_md_utils import set_debug_level

    set_debug_level(directive)

    command, object_type, object_action = extract_command_plus(txt)
    print(Markdown(f"# {command}\n"))

    parsed_output = parse_upsert_command(egeria_client, object_type, object_action, txt, directive)
    attributes = parsed_output['attributes']
    display_name = attributes['Display Name'].get('value', None)
    qualified_name = parsed_output.get('qualified_name', None)
    guid = parsed_output.get('guid', None)
    valid = parsed_output['valid']
    exists = parsed_output['exists']
    output_set = make_format_set_name_from_type(object_type)

    print(Markdown(parsed_output['display']))

    if directive == "display":

        return None
    elif directive == "validate":
        if valid:
            print(Markdown(f"==> Validation of {command} completed successfully!\n"))
        else:
            msg = f"Validation failed for object_action `{command}`\n"
            logger.error(msg)
        return valid

    elif directive == "process":
        logger.debug(json.dumps(parsed_output, indent=4))


        merge_update = attributes.get('Merge Update', {}).get('value', None)

        position = attributes.get('Position', {}).get('value', None)
        min_cardinality = attributes.get('Minimum Cardinality', {}).get('value', None)
        max_cardinality = attributes.get('Maximum Cardinality', {}).get('value', None)

        in_data_structure = attributes.get('In Data Structure', {}).get('value', None)
        data_structure_guid_list = attributes.get('In Data Structure', {}).get('guid_list', None)
        in_data_structure_names = attributes.get('In Data Structure Names', {}).get('name_list', None)

        data_class = attributes.get('Data Class', {}).get('value', None)
        data_class_guid = attributes.get('Data Class', {}).get('guid', None)
        glossary_term = attributes.get('Glossary Term', {}).get('value', None)
        glossary_term_guid = attributes.get('Glossary Term', {}).get('guid', None)
        if glossary_term_guid:
            glossary_term_guid = [glossary_term_guid]

        # name_details_list = attributes.get("dict_list", None)

        in_data_spec = attributes.get("In Data Specification", {}).get("value", None)  # this is a [dict]
        data_spec_name_list = attributes.get("In Data Specification", {}).get("name_list", None)
        data_spec_guid_list = attributes.get("In Data Specification", {}).get("guid_list", None)

        in_data_dictionary = attributes.get('In Data Dictionary', {}).get('value', None)
        in_data_dictionary_names = attributes.get('In Data Dictionary', {}).get('name_list', None)
        data_dict_guid_list = attributes.get("In Data Dictionary", {}).get("guid_list", None)

        parent_data_field = attributes.get('Parent Data Field', {}).get('value', None)
        parent_data_field_guids = attributes.get('Parent Data Field', {}).get('guid_list', None)
        parent_data_field_names = attributes.get('Parent Data Field', {}).get('name_list', None)

        anchor_scope_guid = attributes.get('Anchor Scope GUID', {}).get('value', None)

        replace_all_props = not merge_update

        if not valid:
            if exists and object_action == "Create":
                msg = (f"Create failed because Element `{display_name}` exists - changing `Create` to `Update` in "
                       f"processed output\n\n___")
                logger.error(msg)
                return update_a_command(txt, object_action, object_type, qualified_name, guid)
            else:
                msg = f"Invalid specification - please review\n\n___"
                logger.error(msg)
                return None
        else:
            print(Markdown(f"==> Validation of {command} completed successfully! Proceeding to apply the changes.\n"))

        props_body = set_data_field_body(object_type, qualified_name, attributes)

        try:
            if object_action == "Update":
                if not exists:
                    logger.error(f"Element `{display_name}` does not exist! Updating result document with Create "
                                 f"object_action\n\n___")
                    return update_a_command(txt, object_action, object_type, qualified_name, guid)

                # first update the base data field
                body = set_update_body(object_type, attributes)
                body['properties'] = props_body
                egeria_client.update_data_field(guid, body)
                logger.success(f"Updated  {object_type} `{display_name}` with GUID {guid}")
                # Update data dictionary membership
                update_element_dictionary(qualified_name, {
                    'guid': guid, 'display_name': display_name
                    })
                core_props = egeria_client.get_data_field_by_guid(guid, output_format='MD', report_spec=output_set)  ## update back to by_guid?

                # existing_data_field = egeria_client.get_data_field_by_guid(guid, output_format='JSON')

                # Sync membership in data dictionaries
                update_data_collection_memberships(egeria_client, object_type, data_dict_guid_list, "DataDictionary",
                                                   guid, display_name, replace_all_props)
                logger.success(f"Updating data dictionaries `{in_data_dictionary_names}`")
                core_props += f"\n\n## In Data Dictionary\n\n{in_data_dictionary_names}\n\n"

                # Sync data field related elements (data structure, parent data fields, terms, data classes
                sync_data_field_rel_elements(egeria_client, data_structure_guid_list, parent_data_field_guids,
                                             glossary_term_guid, data_class_guid, guid, display_name, replace_all_props)
                core_props += f"\n\n## In Data Structure {in_data_structure_names}\n\n"
                core_props += f"\n\n## Glossary Term \n\n{glossary_term}\n\n"
                core_props += f"\n\n## Parent Data Field\n\n{parent_data_field_names}\n\n"
                core_props += f"\n\n## Data Class\n\n{data_class}\n\n"
                core_props += "\n_______________________________________________________________________________\n\n"

                # Update data classes
                logger.success(f"Updated Element `{display_name}`\n\n___")
                return core_props

            elif object_action == "Create":
                if valid is False and exists:
                    logger.error(
                        f"\nData Field `{display_name}` already exists and result document updated changing `Create` "
                        f"to `Update` in processed output\n\n___")
                    return update_a_command(txt, object_action, object_type, qualified_name, guid)
                else:
                    # First lets create the data field
                    body = set_create_body(object_type, attributes)
                    body['properties'] = props_body
                    parent_guid = body.get('parentGuid', None)

                    guid = egeria_client.create_data_field(body)
                    if guid:
                        # Now update our element dictionary with the new information
                        update_element_dictionary(qualified_name, {
                            'guid': guid, 'display_name': display_name
                            })
                        # Start assembling the information we will present back out
                        core_props = egeria_client.get_data_field_by_guid(guid, 'MD', report_spec=output_set)

                        # Add the field to any data dictionaries
                        if in_data_dictionary:
                            logger.info(f"Will add to data dictionary `{in_data_dictionary}`")
                            add_member_to_data_collections(egeria_client, data_dict_guid_list, display_name, guid)
                            core_props += f"\n\n## In Data Dictionary\n\n{in_data_dictionary_names}\n\n"

                        # Add the field to any data structures
                        if in_data_structure:
                            core_props += f"\n\n## In Data Structure\n\n{in_data_structure_names}\n\n"
                            for ds_guid in data_structure_guid_list:
                                # todo This is too naive? - need to better accommodate the relationship
                                prop_body = set_rel_prop_body("MemberDataFieldProperties", attributes)
                                prop_body["position"] = position
                                prop_body["minimumCardinality"] = min_cardinality
                                prop_body["maximumCardinality"] = max_cardinality

                                body = set_rel_request_body(object_type, attributes)
                                body['properties'] = prop_body

                                msg = f"Adding field to structure {ds_guid}"
                                logger.info(msg)
                                egeria_client.link_member_data_field(ds_guid, guid, body)
                            core_props += f"\n\n## In Data Structure {in_data_structure_names}\n\n"

                        if glossary_term:
                            if glossary_term_guid:
                                term_body = set_rel_prop_body("GlossaryTermProperties", attributes)
                                body = set_rel_request_body(object_type, attributes)
                                body['properties'] = term_body

                                core_props += f"\n\n## Glossary Term \n\n{glossary_term}\n\n"
                                egeria_client.link_semantic_definition(guid, glossary_term_guid, term_body)

                        if parent_data_field_guids:
                            # parent_df_body = {
                            #     "class": "MetadataSourceRequestBody", "externalSourceGUID": external_source_guid,
                            #     "externalSourceName": external_source_name, "effectiveTime": effective_time,
                            #     "forLineage": for_lineage, "forDuplicateProcessing": for_duplicate_processing
                            #     }

                            # egeria_client.link_nested_data_field(parent_data_field_guid, guid, parent_df_body)
                            for parent_guid in parent_data_field_guids:
                                egeria_client.link_nested_data_field(parent_guid, guid)
                            core_props += f"\n\n## Parent Data Field\n\n{parent_data_field_names}\n\n"

                        # Link data class
                        if data_class:
                            prop_body = set_rel_prop_body("DataClassProperties", attributes)
                            body = set_rel_request_body(object_type, attributes)
                            body['properties'] = prop_body

                            core_props += f"\n\n## Data Class\n\n{data_class}\n\n"
                            egeria_client.link_data_class_definition(guid, data_class_guid, body)
                            msg = f"Adding data class `{data_class}` to data field {display_name}"
                            logger.info(msg)

                        logger.success(f"Created Element `{display_name}` with guid `{guid}`")
                        logger.success("=====================================================\n\n")
                        core_props += "\n___\n\n"
                        return core_props

                    else:
                        logger.error(f"Failed to create Term `{display_name}`\n\n___")
                        return None

        except Exception as e:
            logger.error(f"Error performing {command}: {e}\n\n___")
            return None
    else:
        return None


@logger.catch
def process_data_class_upsert_command(egeria_client: EgeriaTech, txt: str, directive: str = "display") -> Optional[str]:
    """
    Processes a data class create or update object_action by extracting key attributes such as
    spec name, parent_guid, parent_relationship_type, parent_at_end_1, category

    :param txt: A string representing the input cell to be processed for
        extracting glossary-related attributes.
    :param directive: an optional string indicating the directive to be used - display, validate or execute
    :return: A string summarizing the outcome of the processing.
    """

    command, object_type, object_action = extract_command_plus(txt)
    print(Markdown(f"# {command}\n"))

    parsed_output = parse_upsert_command(egeria_client, object_type, object_action, txt, directive)

    attributes = parsed_output['attributes']
    display_name = attributes['Display Name'].get('value', None)
    qualified_name = parsed_output.get('qualified_name', None)
    guid = parsed_output.get('guid', None)
    valid = parsed_output['valid']
    exists = parsed_output['exists']
    output_set = make_format_set_name_from_type(object_type)

    print(Markdown(parsed_output['display']))

    if directive == "display":

        return None
    elif directive == "validate":
        if valid:
            print(Markdown(f"==> Validation of {command} completed successfully!\n"))
        else:
            msg = f"Validation failed for object_action `{command}`\n"
            logger.error(msg)
        return valid

    elif directive == "process":
        logger.debug(json.dumps(parsed_output, indent=4))

        external_source_guid = attributes.get('External Source Name', {}).get('guid', None)
        external_source_name = attributes.get('External Source Name', {}).get('value', None)
        effective_time = attributes.get('Effective Time', {}).get('value', None)
        for_lineage = attributes.get('For Lineage', {}).get('value', False)
        for_duplicate_processing = attributes.get('For Duplicate Processing', {}).get('value', False)
        anchor_guid = attributes.get('Anchor ID', {}).get('guid', None)
        is_own_anchor = attributes.get('Is Own Anchor', {}).get('value', None)
        # parent_id = attributes.get('Parent ID', {}).get('value', None)
        # parent_guid = attributes['Parent ID'].get('guid', None)
        # parent_relationship_type_name = attributes.get('Parent Relationship Type Name', {}).get('value', None)
        # parent_relationship_properties = attributes.get('Parent Relationship Properties',{}).get('value', None)
        # parent_at_end1 = attributes.get('Parent at End1', {}).get('value', None)

        namespace = attributes.get('Namespace', {}).get('value', None)
        description = attributes.get('Description', {}).get('value', None)
        version_id = attributes.get('Version Identifier', {}).get('value', None)

        ###############
        match_property_names = attributes.get('Match Property Names', {}).get('value', [])
        specification_details = attributes.get('Specification Details', {}).get('value', {})
        match_threshold = attributes.get('Match Threshold', {}).get('value', 0)
        specification = attributes.get('Specification', {}).get('value', None)
        data_type = attributes.get('Data Type', {}).get('value', None)
        is_nullable = attributes.get('Is Nullable', {}).get('value', True)
        allow_duplicates = attributes.get('Allow Duplicates', {}).get('value', True)
        default_value = attributes.get('Default Value', {}).get('value', None)
        average_value = attributes.get('Average Value', {}).get('value', None)
        value_list = attributes.get('Value List', {}).get('value', None)
        value_range_from = attributes.get('Value Range From', {}).get('value', None)
        value_range_to = attributes.get('Value Range To', {}).get('value', None)
        sample_values = attributes.get('Sample Values', {}).get('value', [])
        data_patterns = attributes.get('Data Patterns', {}).get('value', [])
        additional_properties = attributes.get('Additional Properties', {}).get('value', {})
        output_set = make_format_set_name_from_type(object_type)

        ###############
        aliases = attributes.get('Aliases', {}).get('value', None)
        name_patterns = attributes.get('Name Patterns', {}).get('value', None)

        min_length = attributes.get('Minimum Length', {}).get('value', None)
        length = attributes.get('Length', {}).get('value', None)
        precision = attributes.get('Precision', {}).get('value', None)
        ordered_values = attributes.get('Ordered Values', {}).get('value', None)
        sort_order = attributes.get('Sort Order', {}).get('value', None)
        effective_from = attributes.get('Effective From', {}).get('value', None)
        effective_to = attributes.get('Effective To', {}).get('value', None)

        glossary_term = attributes.get('Glossary Term', {}).get('value', None)
        glossary_term_guid = attributes.get('Glossary Term', {}).get('guid', None)

        merge_update = attributes.get('Merge Update', {}).get('value', True)

        position = attributes.get('Position', {}).get('value', None)
        min_cardinality = attributes.get('Minimum Cardinality', {}).get('value', None)
        max_cardinality = attributes.get('Maximum Cardinality', {}).get('value', None)

        in_data_structure = attributes.get('In Data Structure', {}).get('value', None)
        data_structure_guid_list = attributes.get('In Data Structure', {}).get('guid_list', None)
        in_data_structure_names = attributes.get('In Data Structure Names', {}).get('name_list', None)

        data_class = attributes.get('Data Class', {}).get('value', None)
        glossary_term = attributes.get('Glossary Term', {}).get('value', None)

        glossary_term_guid = attributes.get('Glossary Term', {}).get('guid', None)

        in_data_dictionary = attributes.get('In Data Dictionary', {}).get('value', None)
        in_data_dictionary_names = attributes.get('In Data Dictionary', {}).get('name_list', None)
        data_dict_guid_list = attributes.get("In Data Dictionary", {}).get("guid_list", None)

        containing_data_class = attributes.get('Containing Data Class', {}).get('value', None)
        containing_data_class_guids = attributes.get('Containing Data Class', {}).get('guid_list', None)
        containing_data_class_names = attributes.get('Containing Data Class', {}).get('name_list', None)

        specializes_data_class = attributes.get('Specializes Data Class', {}).get('value', None)
        specializes_data_class_guid = attributes.get('Specializes Data Class', {}).get('guid', None)
        specializes_data_class_name = attributes.get('Specializes Data Class', {}).get('name', None)

        anchor_scope_guid = attributes.get('Anchor Scope GUID', {}).get('value', None)

        replace_all_props = not merge_update

        if not valid:
            if exists and object_action == "Create":
                msg = (f"Create failed because Element `{display_name}` exists - changing `Create` to `Update` in "
                       f"processed output\n\n___")
                logger.error(msg)
                return update_a_command(txt, object_action, object_type, qualified_name, guid)
            else:
                msg = f"Invalid specification - please review\n\n___"
                return None
        else:
            print(Markdown(f"==> Validation of {command} completed successfully! Proceeding to apply the changes.\n"))

        try:
            if object_action == "Update":
                if not exists:
                    logger.error(f"Element `{display_name}` does not exist! Updating result document with Create "
                                 f"object_action\n\n___")
                    return update_a_command(txt, object_action, object_type, qualified_name, guid)

                # first update the base data class
                body = {
                    "class": "UpdateElementRequestBody", "externalSourceGUID": external_source_guid,
                    "externalSourceName": external_source_name, "effectiveTime": effective_time,
                    "forLineage": for_lineage, "forDuplicateProcessing": for_duplicate_processing, "properties": {
                        "class": "DataClassProperties", "qualifiedName": qualified_name, "displayName": display_name,
                        "description": description, "namespace": namespace, "matchPropertyNames": match_property_names,
                        "matchThreshold": match_threshold, "specification": specification,
                        "specificationDetails": specification_details, "dataType": data_type,
                        "allowsDuplicateValues": allow_duplicates, "isNullable": is_nullable,
                        "defaultValue": default_value, "averageValue": average_value, "valueList": value_list,
                        "valueRangeFrom": value_range_from, "valueRangeTo": value_range_to,
                        "sampleValues": sample_values, "dataPatterns": data_patterns,
                        "additionalProperties": additional_properties
                        }
                    }

                egeria_client.update_data_class(guid, body, not merge_update)
                logger.success(f"Updated  {object_type} `{display_name}` with GUID {guid}")
                # Update data dictionary membership
                update_element_dictionary(qualified_name, {
                    'guid': guid, 'display_name': display_name
                    })
                core_props = egeria_client.get_data_class_by_guid(guid, None, 'MD', report_spec=output_set)

                # Sync membership in data dictionaries
                update_data_collection_memberships(egeria_client, object_type, data_dict_guid_list, "DataDictionary",
                                                   guid, display_name, replace_all_props)
                logger.success(f"Updating data dictionaries `{in_data_dictionary_names}`")
                core_props += f"\n\n## In Data Dictionary\n\n{in_data_dictionary_names}\n\n"

                # Sync data field related elements (data structure, parent data fields, terms, data classes
                sync_data_class_rel_elements(egeria_client, containing_data_class_guids, glossary_term_guid,
                                             specializes_data_class_guid, guid, display_name, replace_all_props)

                core_props += f"\n\n## Glossary Term \n\n{glossary_term}\n\n"
                core_props += f"\n\n## Containing Data Class\n\n{containing_data_class_names}\n\n"
                core_props += "\n___\n\n"

                # Update data classes
                logger.success(f"Updated Element `{display_name}`\n\n___")
                return core_props

            elif object_action == "Create":
                if valid is False and exists:
                    logger.error(
                        f"\nData Class `{display_name}` already exists and result document updated changing `Create` "
                        f"to `Update` in processed output\n\n___")
                    return update_a_command(txt, object_action, object_type, qualified_name, guid)
                else:
                    # First lets create the data class
                    body = {
                        "class": "NewElementRequestBody", "properties": {
                            "class": "DataClassProperties", "qualifiedName": qualified_name,
                            "displayName": display_name, "description": description, "namespace": namespace,
                            "matchPropertyNames": match_property_names, "matchThreshold": match_threshold,
                            "specification": specification, "specificationDetails": specification_details,
                            "dataType": data_type, "allowsDuplicateValues": allow_duplicates, "isNullable": is_nullable,
                            "defaultValue": default_value, "averageValue": average_value, "valueList": value_list,
                            "valueRangeFrom": value_range_from, "valueRangeTo": value_range_to,
                            "sampleValues": sample_values, "dataPatterns": data_patterns,
                            "additionalProperties": additional_properties
                            }
                        }
                    guid = egeria_client.create_data_class(body)
                    if guid:
                        # Now update our element dictionary with the new information
                        update_element_dictionary(qualified_name, {
                            'guid': guid, 'display_name': display_name
                            })
                        # Start assembling the information we will present back out
                        core_props = egeria_client.get_data_class_by_guid(guid, None, 'MD', report_spec=output_set)

                        # Add the field to any data dictionaries
                        if in_data_dictionary:
                            logger.info(f"Will add to data dictionary `{in_data_dictionary}`")
                            add_member_to_data_collections(egeria_client, data_dict_guid_list, display_name, guid)
                            core_props += f"\n\n## In Data Dictionary\n\n{in_data_dictionary_names}\n\n"

                        if glossary_term:
                            if glossary_term_guid:
                                glossary_body = {
                                    "class": "RelationshipRequestBody", "externalSourceGUID": external_source_guid,
                                    "externalSourceName": external_source_name, "effectiveTime": effective_time,
                                    "forLineage": for_lineage, "forDuplicateProcessing": for_duplicate_processing
                                    }

                                core_props += f"\n\n## Glossary Term \n\n{glossary_term}\n\n"
                                egeria_client.link_semantic_definition(guid, glossary_term_guid, glossary_body)

                        if containing_data_class_guids:
                            for dc_guid in containing_data_class_guids:
                                egeria_client.link_nested_data_class(dc_guid, guid)
                            core_props += f"\n\n## Parent Data Field\n\n{containing_data_class_names}\n\n"

                        if specializes_data_class_guid:
                            egeria_client.link_specialist_data_class(specializes_data_class_guid, guid)
                            core_props += f"\n\n## Specialized Data Field\n\n{specializes_data_class_name}\n\n"

                        logger.success(f"Created Element `{display_name}`")
                        core_props += "\n___\n\n"
                        return core_props

                    else:
                        logger.error(f"Failed to create Term `{display_name}`\n\n___")
                        return None

        except Exception as e:
            logger.error(f"Error performing {command}: {e}\n\n___")
            return None
    else:
        return None




#!/usr/bin/env python3
"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.


A object_action line interface for Egeria Data techs.

This is an emerging capability based on the **click** package. Feedback welcome!

"""
import os
import sys

import click
from trogon import tui

from commands.cat.list_tech_types import display_tech_types
from commands.cli.ops_config import Config
from commands.tech.get_element_info import display_elements
from commands.tech.get_guid_info import display_guid
from commands.tech.get_tech_details import tech_details_viewer
from commands.tech.get_tech_type_template import template_viewer
from commands.tech.list_all_om_type_elements import list_elements
from commands.tech.list_all_om_type_elements_x import list_elements_x
from commands.tech.list_all_related_elements import list_related_elements
from commands.tech.list_anchored_elements import display_anchored_elements
from commands.tech.list_asset_types import display_asset_types
from commands.tech.list_elements_by_classification_by_property_value import (
    find_elements_by_classification_by_prop_value,
)
from commands.tech.list_elements_by_property_value import (
    find_elements_by_prop_value,
)
from commands.tech.list_elements_by_property_value_x import (
    find_elements_by_prop_value_x,
)
from commands.tech.list_elements_for_classification import (
    list_classified_elements,
)
from commands.tech.list_gov_action_processes import display_gov_processes
from commands.tech.list_information_supply_chains import supply_chain_viewer
from commands.tech.list_registered_services import display_registered_svcs
from commands.tech.list_related_elements_with_prop_value import (
    list_related_elements_with_prop_value,
)
from commands.tech.list_related_specification import (
    display_related_specification,
)
from commands.tech.list_relationship_types import display_relationship_types
from commands.tech.list_solution_blueprints import blueprint_list
from commands.tech.list_solution_components import solution_component_list
from commands.tech.list_solution_roles import solution_role_list
from commands.tech.list_tech_templates import display_templates_spec
from commands.tech.list_valid_metadata_values import display_metadata_values
from commands.tech.generic_actions import delete_element

@tui()
@click.version_option("0.0.1", prog_name="egeria_ops")
@click.group()
@click.option(
    "--server",
    default=os.environ.get("EGERIA_METADATA_STORE", "qs-metadata-store"),
    help="Egeria metadata store to work with",
)
@click.option(
    "--url",
    default=os.environ.get("EGERIA_PLATFORM_URL", "https://localhost:9443"),
    help="URL of Egeria metadata store platform to connect to",
)
@click.option(
    "--integration_daemon",
    default=os.environ.get("EGERIA_INTEGRATION_DAEMON", "qs-integration-daemon"),
    help="Egeria integration daemon to work with",
)
@click.option(
    "--integration_daemon_url",
    default=os.environ.get("EGERIA_INTEGRATION_DAEMON_URL", "https://localhost:9443"),
    help="URL of Egeria integration daemon platform to connect to",
)
@click.option(
    "--view_server",
    default=os.environ.get("EGERIA_VIEW_SERVER", "qs-view-server"),
    help="Egeria view server to work with",
)
@click.option(
    "--view_server_url",
    default=os.environ.get("EGERIA_VIEW_SERVER_URL", "https://localhost:9443"),
    help="URL of Egeria view server platform to connect to",
)
@click.option(
    "--engine_host",
    default=os.environ.get("EGERIA_ENGINE_HOST", "qs-engine-host"),
    help="Egeria engine host to work with",
)
@click.option(
    "--engine_host_url",
    default=os.environ.get("EGERIA_ENGINE_HOST_URL", "https://localhost:9443"),
    help="URL of Egeria engine host platform to connect to",
)
@click.option(
    "--admin_user",
    default=os.environ.get("EGERIA_ADMIN_USER", "garygeeke"),
    help="Egeria admin user",
)
@click.option(
    "--admin_user_password",
    default=os.environ.get("EGERIA_ADMIN_PASSWORD", "secret"),
    help="Egeria admin password",
)
@click.option(
    "--userid",
    default=os.environ.get("EGERIA_USER", "peterprofile"),
    help="Egeria user",
)
@click.option(
    "--password",
    default=os.environ.get("EGERIA_USER_PASSWORD", "secret"),
    help="Egeria user password",
)
@click.option("--timeout", default=60, help="Number of seconds to wait")
@click.option(
    "--jupyter",
    is_flag=True,
    default=os.environ.get("EGERIA_JUPYTER", "False"),
    help="Enable for rendering in a Jupyter terminal",
)
@click.option(
    "--width",
    default=os.environ.get("EGERIA_WIDTH", "200"),
    help="Screen width, in characters, to use",
)
@click.option(
    "--home_glossary_guid",
    default=os.environ.get("EGERIA_HOME_GLOSSARY_GUID", None),
    help="Glossary guid to use as the home glossary",
)
@click.option(
    "--glossary_path",
    default=os.environ.get("EGERIA_GLOSSARY_PATH", "/home/jovyan/loading-bay/glossary"),
    help="Path to glossary import/export files",
)
@click.pass_context
def cli(
    ctx,
    server,
    url,
    view_server,
    view_server_url,
    integration_daemon,
    integration_daemon_url,
    engine_host,
    engine_host_url,
    admin_user,
    admin_user_password,
    userid,
    password,
    timeout,
    jupyter,
    width,
    home_glossary_guid,
    home_glossary_path,
):
    """An Egeria Command Line interface for Operations"""
    ctx.obj = Config(
        server,
        url,
        view_server,
        view_server_url,
        integration_daemon,
        integration_daemon_url,
        engine_host,
        engine_host_url,
        admin_user,
        admin_user_password,
        userid,
        password,
        timeout,
        jupyter,
        width,
        home_glossary_guid,
        home_glossary_path,
    )
    ctx.max_content_width = 200
    ctx.ensure_object(Config)


@cli.group("show")
@click.pass_context
def show(ctx):
    """Display an Egeria Object"""
    pass


@show.group("info")
@click.pass_context
def show_info(ctx):
    """Show various Egeria information"""
    pass


@show.group("tech-types")
@click.pass_context
def show_tech(ctx):
    """Show information about Egeria technology types"""
    pass


@show.group("elements")
@click.pass_context
def show_elements(ctx):
    """Show information about Egeria elements"""


@show.group("solution-architecture")
@click.pass_context
def show_supply_chains(ctx):
    """Show information about Information Supply Chains"""


@show_elements.command("guid-info")
@click.argument("guid", nargs=1)
@click.pass_context
def show_guid_info(ctx, guid):
    """Display guid information

    Usage: show guid-info <a guid>

    """
    c = ctx.obj
    display_guid(guid, c.server, c.url, c.userid, c.password, c.jupyter, c.width)


@show_elements.command("anchored_elements")
@click.pass_context
@click.option(
    "--property_value",
    default="DeployedDatabaseSchema",
    help="value we are searching for",
)
@click.option(
    "--prop-list",
    default="anchorTypeName",
    help="List of properties we are searching",
)
def list_anchored_elements(ctx, property_value: str, prop_list: str):
    """List anchored elements with the specified properties"""
    c = ctx.obj
    if type(prop_list) is str:
        property_names = prop_list.split(",")
    elif type(prop_list) is list:
        property_names = prop_list
    else:
        property_names = []
        print("Error --> Invalid property list - must be a string or list")
        sys.exit(4)
    display_anchored_elements(
        property_value,
        [property_names],
        c.view_server,
        c.view_server_url,
        c.userid,
        c.password,
        c.timeout,
        c.jupyter,
        c.width,
    )


@show_elements.command("elements-by-classification")
@click.option(
    "--om-type",
    default="Referenceable",
    help="Open Metadata type to filter by",
)
@click.option(
    "--classification",
    default="GovernanceProject",
    help="Classification to filter byt",
)
@click.pass_context
def show_elements_by_classification(ctx, om_type, classification):
    """Show elements by classification"""
    c = ctx.obj
    list_classified_elements(
        om_type,
        classification,
        c.view_server,
        c.view_server_url,
        c.userid,
        c.password,
        c.jupyter,
        c.width,
    )


@show_elements.command("elements-by-classification-by-prop-value")
@click.option(
    "--classification",
    default="GovernanceProject",
    help="Classification to filter by",
)
@click.option(
    "--property_value",
    help="Property value to filter by",
)
@click.option(
    "--property_names",
    help="List of properties to search by",
)
@click.option(
    "--om-type",
    default="Referenceable",
    help="Open Metadata type to filter by",
)
@click.pass_context
def show_elements_by_classification_by_prop(
    ctx, classification, property_value, property_names, om_type
):
    """Show elements by classification and property value"""
    c = ctx.obj
    find_elements_by_classification_by_prop_value(
        om_type,
        classification,
        property_value,
        [property_names],
        c.view_server,
        c.view_server_url,
        c.userid,
        c.password,
        c.jupyter,
        c.width,
    )


@show_elements.command("elements-by-prop-value")
@click.option(
    "--property_value",
    help="Property value to filter by",
)
@click.option(
    "--property_names",
    help="List of properties to search by",
)
@click.option(
    "--om-type",
    default="Referenceable",
    help="Open Metadata type to filter by",
)
@click.option(
    "--extended",
    is_flag=True,
    default=False,
    help="If True, feedback information is displayed",
)
@click.pass_context
def show_elements_by_classification_by_prop(
    ctx, property_value, property_names, om_type, extended
):
    """Show elements by classification and property value"""
    c = ctx.obj
    if extended:
        find_elements_by_prop_value_x(
            om_type,
            property_value,
            [property_names],
            c.view_server,
            c.view_server_url,
            c.userid,
            c.password,
            c.jupyter,
            c.width,
        )
    else:
        find_elements_by_prop_value(
            om_type,
            property_value,
            [property_names],
            c.view_server,
            c.view_server_url,
            c.userid,
            c.password,
            c.jupyter,
            c.width,
        )


@show_elements.command("related-elements")
@click.option(
    "--element-guid",
    help="GUID of the Element to navigate from.",
)
@click.option(
    "--om-type",
    default="Referenceable",
    help="Open metadata type to filter by.",
)
@click.option(
    "--rel-type",
    default="Certification",
    help="Relationship type to follow.",
)
@click.pass_context
def show_related_elements(ctx, element_guid, om_type, rel_type):
    """Show all elements related to specified guid"""
    c = ctx.obj
    list_related_elements(
        element_guid,
        om_type,
        rel_type,
        c.view_server,
        c.view_server_url,
        c.userid,
        c.password,
        c.jupyter,
        c.width,
    )


@show_elements.command("related-elements_by_prop")
@click.option(
    "--element-guid",
    help="GUID of the Element to navigate from.",
)
@click.option(
    "--rel-type",
    default="Certification",
    help="Relationship type to follow.",
)
@click.option(
    "--property_value",
    help="Property value to filter by",
)
@click.option(
    "--property_names",
    help="List of properties to search by",
)
@click.option(
    "--om-type",
    default="Referenceable",
    help="Open metadata type to filter by.",
)
@click.pass_context
def show_related_elements(
    ctx, element_guid, rel_type, property_value, property_names, om_type
):
    """Show elements related to specified guid and property value"""
    c = ctx.obj
    list_related_elements_with_prop_value(
        element_guid,
        rel_type,
        property_value,
        [property_names],
        om_type,
        c.view_server,
        c.view_server_url,
        c.userid,
        c.password,
        c.jupyter,
        c.width,
    )


@show_elements.command("related-specifications")
@click.pass_context
@click.argument("element-guid")
def show_related_specifications(ctx, element_guid):
    """List specifications related to the given Element"""
    c = ctx.obj
    display_related_specification(
        element_guid,
        c.view_server,
        c.view_server_url,
        c.userid,
        c.password,
        c.jupyter,
        c.width,
    )


@show_tech.command("tech-types")
@click.option("--search-string", default="*", help="Tech type to search for")
@click.pass_context
def show_tech_types(ctx, search_string):
    """List deployed technology types

    Usage: show tech-types <optional search-string>

    All tech-types will be returned if no search-string is specified.

    """

    c = ctx.obj
    display_tech_types(
        search_string, c.view_server, c.view_server_url, c.userid, c.password
    )


@show_tech.command("tech-details")
@click.argument("tech-name")
@click.pass_context
def show_tech_details(ctx, tech_name):
    """Display a live status view of Egeria servers for the specified Egeria platform

    Usage: show tech-details <tech-name>

           tech-name is a valid technology name (see 'show tech-types')
    """
    c = ctx.obj
    tech_details_viewer(
        tech_name,
        c.view_server,
        c.view_server_url,
        c.userid,
        c.password,
        c.jupyter,
        c.width,
    )


@show_tech.command("tech-type-templates")
@click.option(
    "--tech-type",
    default="PostgreSQL Server",
    help="Specific tech type to get elements for",
)
@click.pass_context
def show_tech_type_templates(ctx, tech_type):
    """List technology type templates"""
    c = ctx.obj
    template_viewer(
        tech_type,
        c.view_server,
        c.view_server_url,
        c.userid,
        c.password,
        c.jupyter,
        c.width,
    )


@show_info.command("asset-types")
@click.pass_context
def show_asset_types(ctx):
    """Display known asset types"""
    c = ctx.obj
    display_asset_types(
        c.view_server, c.view_server_url, c.userid, c.password, c.jupyter, c.width
    )


@show_info.command("registered-services")
@click.option(
    "--services",
    type=click.Choice(
        [
            "all",
            "access-services",
            "common-services",
            "engine-services",
            "governance-services",
            "integration-services",
            "view-services",
        ],
        case_sensitive=False,
    ),
    default="all",
    help="Which service group to display",
)
@click.pass_context
def show_registered_services(ctx, services):
    """Show information about a registered services"""
    c = ctx.obj
    display_registered_svcs(
        services,
        c.view_server,
        c.view_server_url,
        c.userid,
        c.password,
        c.jupyter,
        c.width,
    )


@show_info.command("relationship-types")
@click.option(
    "--om-type",
    default="Referenceable",
    help="Relationship type to get information about",
)
@click.pass_context
def show_relationship_types(ctx, rel_type):
    """Show information about the specified relationship type"""
    c = ctx.obj
    display_relationship_types(
        rel_type,
        c.view_server,
        c.view_server_url,
        c.userid,
        c.password,
        False,
        c.jupyter,
        c.width,
    )


@show_tech.command("tech-templates")
@click.pass_context
@click.option(
    "--search-string", default="*", help="Technology type to get information about"
)
def tech_templates(ctx, search_string):
    """Display template information about the specified technology."""
    c = ctx.obj
    template_viewer(
        search_string,
        c.view_server,
        c.view_server_url,
        c.userid,
        c.password,
        c.jupyter,
        c.width,
    )


@show_tech.command("tech-template-spec")
@click.pass_context
@click.option(
    "--search-string", default="*", help="Technology type to get information about"
)
def tech_template_spec(ctx, search_string):
    """Display template specification information about the specified technology."""
    c = ctx.obj
    display_templates_spec(
        search_string,
        c.view_server,
        c.view_server_url,
        c.userid,
        c.password,
        c.jupyter,
        c.width,
    )


@show_info.command("gov-action-processes")
@click.pass_context
@click.option("--search-string", default="*", help="Search string")
def gov_action_processes(ctx, search_string):
    """Display available governance action processes."""
    c = ctx.obj
    display_gov_processes(
        search_string,
        c.view_server,
        c.view_server_url,
        c.userid,
        c.password,
        c.jupyter,
        c.width,
    )


@show_info.command("valid-metadata-values")
@click.pass_context
@click.option("--property", default="projectHealth", help="Metadata property to query")
@click.option("--type-name", default="Project", help="Metadata type to query")
def valid_metadata_values(ctx, property, type_name):
    """Display the valid metadata values for a property and type"""
    c = ctx.obj
    display_metadata_values(
        property,
        type_name,
        c.view_server,
        c.view_server_url,
        c.userid,
        c.password,
        False,
        c.jupyter,
        c.width,
    )


@show_info.command("processes")
@click.pass_context
def list_element_info(ctx):
    """Display processes"""
    c = ctx.obj
    list_elements(
        "GovernanceActionProcess",
        c.view_server,
        c.view_server_url,
        c.userid,
        c.password,
        c.jupyter,
        c.width,
    )


@show_elements.command("get-elements")
@click.pass_context
@click.option("--om_type", default="Referenceable", help="Metadata type to query")
def get_element_info(ctx, om_type):
    """Display graph of elements for an Open Metadata Type"""
    c = ctx.obj
    display_elements(
        om_type,
        c.view_server,
        c.view_server_url,
        c.userid,
        c.password,
        c.jupyter,
        c.width,
    )


@show_elements.command("elements")
@click.pass_context
@click.option(
    "--extended",
    is_flag=True,
    default=False,
    help="If True, feedback information is displayed",
)
@click.option("--om_type", default="Referenceable", help="Metadata type to query")
def list_all_om_type_elements(ctx, om_type, extended):
    """Display all elements of a specific Open Metadata Type"""
    c = ctx.obj
    if extended:
        list_elements_x(
            om_type,
            c.view_server,
            c.view_server_url,
            c.userid,
            c.password,
            c.jupyter,
            c.width,
        )
    else:
        list_elements(
            om_type,
            c.view_server,
            c.view_server_url,
            c.userid,
            c.password,
            c.jupyter,
            c.width,
        )


@show_supply_chains.command("supply-chains")
@click.option("--search-string", default="*", help="Search string")
@click.pass_context
def list_supply_chains(ctx, search_string):
    """Display supply chains"""
    c = ctx.obj
    supply_chain_viewer(
        search_string,
        c.view_server,
        c.view_server_url,
        c.userid,
        c.password,
        c.jupyter,
        c.width,
    )


@show_supply_chains.command("blueprints")
@click.option("--search-string", default="*", help="Search string")
@click.pass_context
def list_blueprints(ctx, search_string):
    """Display solution blueprints"""
    c = ctx.obj
    blueprint_list(
        search_string,
        c.view_server,
        c.view_server_url,
        c.userid,
        c.password,
        c.jupyter,
        c.width,
    )


@show_supply_chains.command("solution-roles")
@click.option("--search-string", default="*", help="Search string")
@click.pass_context
def list_solution_roles(ctx, search_string):
    """Display solution roles"""
    c = ctx.obj
    solution_role_list(
        search_string,
        c.view_server,
        c.view_server_url,
        c.userid,
        c.password,
        c.jupyter,
        c.width,
    )


@show_supply_chains.command("solution-components")
@click.option("--search-string", default="*", help="Search string")
@click.pass_context
def list_solution_components(ctx, search_string):
    """Display solution componentss"""
    c = ctx.obj
    solution_component_list(
        search_string,
        c.view_server,
        c.view_server_url,
        c.userid,
        c.password,
        c.jupyter,
        c.width,
    )


#
#  Tell
#


@cli.group("tell")
@click.pass_context
def tell(ctx):
    """Perform actions an Egeria Objects"""
    pass
tell.add_command(delete_element)

if __name__ == "__main__":
    cli()



"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.



This module is for testing the metadata explorer view service class and methods.
The routines assume that pytest is being used as the test tool and framework.

A running Egeria environment is needed to run these tests.

"""

import json
import time

from rich import print, print_json

from pyegeria import SolutionArchitect, PyegeriaException
from pyegeria._exceptions import (
    InvalidParameterException,
    PropertyServerException,
    UserNotAuthorizedException,
    print_exception_response,
)
from pyegeria._exceptions_new import PyegeriaException, print_basic_exception

disable_ssl_warnings = True


def jprint(info, comment=None):
    if comment:
        print(comment)
    print(json.dumps(info, indent=2))


def valid_guid(guid):
    if (guid is None) or (type(guid) is not str):
        return False
    else:
        return True


class TestSolutionArchitect:
    good_view_server_1 = "view-server"
    platform_url = "https://localhost:9443"
    view_server = "qs-view-server"
    user = "erinoverview"
    password = "secret"

    #
    ##
    #


    def test_create_info_supply_chains(self):
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
                )
            s_client.create_egeria_bearer_token()
            display_name = "puddys_supply_chain"
            q_name = s_client.__create_qualified_name__("InformationSupplyChain",display_name)
            # body = {
            #     "class": "NewElementRequestBody",
            #     "isOwnAnchor": True,
            #     "properties": {
            #         "class": "InformationSupplyChainProperties",
            #         "displayName": display_name,
            #         "qualifiedName": q_name
            #         }
            #     }
            body ={
                'class': 'NewElementRequestBody',
                'anchorGUID': None,
                'anchorScopeGUID': None,
                'effectiveTime': None,
                'externalSourceGUID': None,
                'externalSourceName': None,
                'forDuplicateProcessing': False,
                'forLineage': False,
                'isOwnAnchor': True,
                'parentAtEnd1': True,
                'parentGUID': None,
                'parentRelationshipTypeName': None,
                'properties': {
                    'class': 'InformationSupplyChainProperties',
                    'additionalProperties': None,
                    'description': 'My own personal supply chain',
                    'displayName': display_name,
                    'effectiveFrom': None,
                    'effectiveTo': None,
                    'extendedProperties': None,
                    'purposes': ['hegemony', 'betterment of mankind', 'pranks'],
                    'qualifiedName': q_name,
                    'scope': 'universal',
                    'version': None
                    }
                }
            start_time = time.perf_counter()
            response = s_client.create_info_supply_chain(body)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
                )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
                InvalidParameterException,
                PropertyServerException,
                UserNotAuthorizedException,
                ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_update_info_supply_chains(self):
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
                )
            s_client.create_egeria_bearer_token()

            body = {
              "class" : "UpdateElementRequestBody",
              "properties": {
                  "class": "InformationSupplyChainProperties",
                  "description": "updated description",
                  "scope": "updated scope"
              }
            }
            start_time = time.perf_counter()
            guid = "261171b5-9742-41d8-af6d-d68e98f05141"
            s_client.update_info_supply_chain(guid,body, False)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds"
                )

            assert True
        except (
                InvalidParameterException,
                PropertyServerException,
                UserNotAuthorizedException,
                ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_delete_info_supply_chains(self):
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
                )
            s_client.create_egeria_bearer_token()
            guid = "261171b5-9742-41d8-af6d-d68e98f05141"
            # body = {
            #     "class": "NewInformationSupplyChainRequestBody",
            #     "properties": {
            #         "displayName": display_name,
            #         "qualifiedName": q_name
            #         }
            #     }
            start_time = time.perf_counter()
            s_client.delete_info_supply_chain(guid)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds"
                )


            assert True
        except (
                InvalidParameterException,
                PropertyServerException,
                UserNotAuthorizedException,
                ) as e:
            print_exception_response(e)
            # if e.exception_error_message_id == 'OMAG-REPOSITORY-HANDLER-404-007':
            #     print("The GUID does not exist")
            #     assert True
            #     return
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_find_information_supply_chains(self):
        filter_string = "puddys_"
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            start_time = time.perf_counter()
            response = s_client.find_information_supply_chains(filter_string, output_format='DICT')
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
            )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
            PyegeriaException
        ) as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_get_information_supply_chain_by_guid(self):
        guid = "7cbcb2ae-0b06-4dc6-83b7-d5951c435fe3"
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            start_time = time.perf_counter()
            response = s_client.get_info_supply_chain_by_guid(guid, output_format='DICT')
            duration = time.perf_counter() - start_time
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
            )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
            PyegeriaException
        ) as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()


    def test_create_complex_chain(self):
        top_chain1 = 'top1'
        top_chain2 = 'top2'
        sub_chain1 = 'sub1'
        sub_chain2 = 'sub2'

        # create chain1 composed of the two sub chains
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
                )
            s_client.create_egeria_bearer_token()
            q_name = s_client.__create_qualified_name__("InformationSupplyChain", top_chain1)
            body = {
                "class": "NewElementRequestBody",
                "isOwnAnchor": True,
                "properties": {
                    "class": "InformationSupplyChainProperties",
                    "displayName": top_chain1,
                    "qualifiedName": q_name
                    }
                }
            guid1 = s_client.create_info_supply_chain(body)

            q_name = s_client.__create_qualified_name__("InformationSupplyChain", top_chain2)
            body = {
                "class": "NewElementRequestBody",
                "isOwnAnchor": True,
                "properties": {
                    "class": "InformationSupplyChainProperties",
                    "displayName": top_chain2,
                    "qualifiedName": q_name
                    }
                }
            guid2 = s_client.create_info_supply_chain(body)

            print(f"Top1: {guid1}, Top2: {guid2}")


        ## put in links

            assert True
        except (
                InvalidParameterException,
                PropertyServerException,
                UserNotAuthorizedException,
                ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_get_information_supply_chain_by_name(self):
        name = "InformationSupplyChain::Personalized Treatment Ordering"
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            start_time = time.perf_counter()
            response = s_client.get_info_supply_chain_by_name(name)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
            )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
            InvalidParameterException,
            PropertyServerException,
            UserNotAuthorizedException,
        ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()


    def test_find_information_supply_chains_body(self):
        filter = "puddy"
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            body = {
                "class": "SearchStringRequestBody",
                # "effective_time": None,
                # "limitResultsByStatus": ["ACTIVE"],
                "sequencingOrder": None,
                "sequencingProperty": None,
                "searchString": filter
                }
            start_time = time.perf_counter()
            response = s_client.find_information_supply_chains(filter,
                                                               body=body, output_format="DICT")
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
            )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
            PyegeriaException
        ) as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_find_all_information_supply_chains(self):

        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            start_time = time.perf_counter()
            response = s_client.find_all_information_supply_chains(output_format='DICT')
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
            )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
            PyegeriaException,
        ) as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_link_peer_supply_chains(self):
        name = "sell to the highest bidder2"
        guid = "fdbedd27-2332-4fe5-b50c-9007238fd527"
        guid2 = "70e1ae07-fc8d-49c1-89de-582f671d2b65"
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            q_name = s_client.__create_qualified_name__("InformationSupplyChainSegment",name)
            body = {
                "class": "RelationshipRequestBody",
                "properties": {
                    "class": "InformationSupplyChainLinkProperties",
                    "label": "a first label",
                    "description": "label description"
                    }
                }


            start_time = time.perf_counter()
            s_client.link_peer_info_supply_chain(guid, guid2, body)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds,"
            )

            assert True
        except (
            InvalidParameterException,
            PropertyServerException,
            UserNotAuthorizedException,
        ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()


    def test_unlink_peer_supply_chains(self):
        segment1_guid = "fdbedd27-2332-4fe5-b50c-9007238fd527"
        segment2_guid = "fdbedd27-2332-4fe5-b50c-9007238fd527"
        body = {
             "class" : "MetadataSourceRequestBody",
              # "externalSourceGUID": "add guid here",
              # "externalSourceName": "add qualified name here",
              # "effectiveTime" : "{{$isoTimestamp}}",
              # "forLineage" : false,
              # "forDuplicateProcessing" : false,

            }
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            start_time = time.perf_counter()


            s_client.unlink_peer_info_supply_chains(segment1_guid,segment2_guid)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds"
            )

            assert True
        except (
            InvalidParameterException,
            PropertyServerException,
            UserNotAuthorizedException,
        ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_compose_info_supply_chain(self):
        guid1 = "8d7c39bc-f2c5-48df-8f34-c76dbe22b376"
        guid2 = "65ef0367-e69b-40e1-b6e8-4a00c12c06fc"
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            body = {
          "class" : "RelationshipRequestBody",
          "properties": {
              "class": "InformationSupplyChainLinkProperties",
            "label": "sub1",
            # "description": "checking to see it works",
          }
        }

            start_time = time.perf_counter()
            s_client.compose_info_supply_chains(guid1,guid2)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds"
            )

            assert True
        except (
            InvalidParameterException,
            PropertyServerException,
            UserNotAuthorizedException,
        ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_decompose_info_supply_chain(self):
        guid1 = "8d7c39bc-f2c5-48df-8f34-c76dbe22b376"
        guid2 = "65ef0367-e69b-40e1-b6e8-4a00c12c06fc"
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            body = {
          "class" : "MetadataSourceRequestBody",
            }


            start_time = time.perf_counter()
            s_client.decompose_info_supply_chains(guid1,guid2)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds"
            )

            assert True
        except (
            InvalidParameterException,
            PropertyServerException,
            UserNotAuthorizedException,
        ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()


    def test_create_solution_blueprint(self):
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
                )
            s_client.create_egeria_bearer_token()
            display_name = "my_second_blueprint"
            q_name = s_client.__create_qualified_name__("Blueprint",display_name)
            body = {
                "class": "NewElementRequestBody",
                "isOwnAnchor": True,
                "properties": {
                    "class": "SolutionBlueprintProperties",
                    "displayName": display_name,
                    "qualifiedName": q_name,
                    "description": "some description"
                    },
                "initialStatus": "ACTIVE"
                }
            start_time = time.perf_counter()
            response = s_client.create_solution_blueprint(body)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
                )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
                PyegeriaException
                ) as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_update_solution_blueprint(self):
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
                )
            s_client.create_egeria_bearer_token()

            body = {
              "class" : "UpdateSolutionBlueprintRequestBody",
              "effectiveTime" : "{{$isoTimestamp}}",
              "properties": {
                "typeName" : "string",
                "class:": "SolutionBlueprintProperties",
                "qualifiedName": "SolnBlueprint::dr egerias blueprint",
                # "description": "updated description4",
                # "displayName": "dr egerias blueprint",
                # "version": "2"
              }
            }

            start_time = time.perf_counter()
            guid = '8e42075c-4bbe-47fb-9e96-8bd60c41bf6d'
            s_client.update_solution_blueprint(guid,body, False)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds"
                )

            assert True
        except (
                InvalidParameterException,
                PropertyServerException,
                UserNotAuthorizedException,
                ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_delete_solution_blueprint(self):
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
                )
            s_client.create_egeria_bearer_token()
            guid = "8e42075c-4bbe-47fb-9e96-8bd60c41bf6d"
            # body = {
            #     "class": "NewInformationSupplyChainRequestBody",
            #     "properties": {
            #         "displayName": display_name,
            #         "qualifiedName": q_name
            #         }
            #     }
            start_time = time.perf_counter()
            s_client.delete_solution_blueprint(guid)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, "
                )

            assert True
        except (
                InvalidParameterException,
                PropertyServerException,
                UserNotAuthorizedException,
                ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_get_solution_blueprint_by_guid(self):
        guid = "c83c3bc4-5858-4267-bb24-d93c5517b215"
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            start_time = time.perf_counter()
            response = s_client.get_solution_blueprint_by_guid(guid, output_format='DICT')
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
            )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
            InvalidParameterException,
            PropertyServerException,
            UserNotAuthorizedException,
        ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()


    def test_get_solution_blueprints_by_name(self):
        name = "dr egerias blueprint"
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            start_time = time.perf_counter()
            response = s_client.get_solution_blueprints_by_name(name)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
            )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
            InvalidParameterException,
            PropertyServerException,
            UserNotAuthorizedException,
        ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()


    def test_find_solution_blueprints(self):
        search_string = "*"
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            start_time = time.perf_counter()
            response = s_client.find_solution_blueprints(search_string, output_format='DICT', report_spec='Solution-Blueprint')
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
            )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
            PyegeriaException
        ) as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_find_all_blueprints(self):
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            start_time = time.perf_counter()
            response = s_client.find_all_solution_blueprints(output_format="DICT")
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds Type: {type(response) if response  else ''}, "
                f" Element count is {len(response) if response else ''}"
            )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
            InvalidParameterException,
            PropertyServerException,
            UserNotAuthorizedException,
        ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_create_solution_roles(self):
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
                )
            body = {
                "class": "NewActorRoleRequestBody",
                "properties": {
                    "class" : "ActorRoleProperties",
                    "typeName": "GovernanceRole",
                    "domainIdentifier": 0,
                    "qualifiedName": "role::trial-manager",
                    "displayName": "Trial Manager",
                    "description": "bureaucrat",
                    "roleId": "trial-manager",
                    "scope": "coco",
                    "title": "Demigod"
                    }
                }
            s_client.create_egeria_bearer_token()

            start_time = time.perf_counter()
            response = s_client.create_solution_role(body)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
                )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
                InvalidParameterException,
                PropertyServerException,
                UserNotAuthorizedException,
                ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_update_solution_role(self):
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
                )
            s_client.create_egeria_bearer_token()

            body = {
              "class" : "UpdateActorRoleRequestBody",
              "properties": {
                "class:": "ActorRoleProperties",
                "description": "updated description - moo",

              }
            }
            start_time = time.perf_counter()
            guid = "4454aeb9-cfaf-4ea0-afa5-b3af3945f341"
            s_client.update_solution_blueprint(guid,body, False)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds"
                )

            assert True
        except (
                InvalidParameterException,
                PropertyServerException,
                UserNotAuthorizedException,
                ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_delete_solution_role(self):
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
                )
            s_client.create_egeria_bearer_token()
            guid = "1f71e403-1187-4f03-a1dd-ae7dc105f06f"
            # body = {
            #     "class": "NewInformationSupplyChainRequestBody",
            #     "properties": {
            #         "displayName": display_name,
            #         "qualifiedName": q_name
            #         }
            #     }
            start_time = time.perf_counter()
            response = s_client.delete_solution_role(guid)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
                )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
                InvalidParameterException,
                PropertyServerException,
                UserNotAuthorizedException,
                ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_get_solution_role_by_guid(self):
        guid = "09e21ac1-39d0-4df2-969c-e41fb29a067d"
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            start_time = time.perf_counter()
            response = s_client.get_solution_role_by_guid(guid, output_format='JSON')
            duration = time.perf_counter() - start_time
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
            )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
            InvalidParameterException,
            PropertyServerException,
            UserNotAuthorizedException,
        ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()


    def test_get_solution_role_by_name(self):
        name = "Clinical Trial Sponsor"
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            start_time = time.perf_counter()
            response = s_client.get_solution_roles_by_name(name)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
            )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
            InvalidParameterException,
            PropertyServerException,
            UserNotAuthorizedException,
        ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()


    def test_find_solution_roles(self):
        search_string = "ClinicalTrialParticipatingHospital"
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            start_time = time.perf_counter()
            response = s_client.find_solution_roles(search_string, output_format='JSON')
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
            )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: \n" + response)

            assert True
        except (
            InvalidParameterException,
            PropertyServerException,
            UserNotAuthorizedException,
        ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_find_all_solution_roles(self):
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            start_time = time.perf_counter()
            response = s_client.find_all_solution_roles(output_format='DICT')
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
            )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
            InvalidParameterException,
            PropertyServerException,
            UserNotAuthorizedException,
        ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()


    def test_create_solution_component(self):
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
                )

            s_client.create_egeria_bearer_token()

            body = {
              "class": "NewElementRequestBody",
              "parentAtEnd1": True,
              "properties": {
                "class": "SolutionComponentProperties",
                "qualifiedName": "SolnComponent::Lab Process::2",
                "displayName": "Lab Processes",
                "description": "a description",
                "solutionComponentType": "dan-test",
                "plannedDeployedImplementationType": "something"
              }
            }


            start_time = time.perf_counter()
            response = s_client.create_solution_component(body)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
                )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
                InvalidParameterException,
                PropertyServerException,
                UserNotAuthorizedException,
                ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_update_solution_component(self):
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
                )
            s_client.create_egeria_bearer_token()

            body = {
              "class" : "UpdateElementRequestBody",
              "mergeUpdate": True,
              "properties": {
                "class": "SolutionComponentProperties",
                "description": "Milvus now",
                "solutionComponentType": "SoftwareServer",
                "plannedDeployedImplementationType": "DataStore"
                  }
            }
            start_time = time.perf_counter()
            guid = '6948894a-5fd4-41e7-a250-955cfede5147'
            s_client.update_solution_component(guid,body)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds"
                )

            assert True
        except (
                PyegeriaException,
                ) as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_delete_solution_component(self):
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
                )
            s_client.create_egeria_bearer_token()
            guid = "9e0ed1ce-00da-45ab-8bb2-a9f47d9598a3"

            start_time = time.perf_counter()
            s_client.delete_solution_component(guid)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds"
                )

            assert True
        except (
                InvalidParameterException,
                PropertyServerException,
                UserNotAuthorizedException,
                ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_link_solution_component_to_blueprint(self):
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
                )
            s_client.create_egeria_bearer_token()
            blueprint = "8a222c5d-b206-454f-b861-2b803cfe3cbd"
            comp = "d026fbd6-709b-45d4-beec-f7a08764b5e2"


            start_time = time.perf_counter()
            guid = "0f8fea0d-d6e7-46e4-a32f-486899868bc0"
            s_client.link_solution_component_to_blueprint(blueprint,comp,None)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds"
                )

            assert True
        except (
                InvalidParameterException,
                PropertyServerException,
                UserNotAuthorizedException,
                ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_link_solution_components(self):
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
                )
            s_client.create_egeria_bearer_token()
            comp1 = "b9a2735d-3a4a-4208-afe8-974e4369e733"
            comp2 = "7186fd6e-1314-47e3-bddf-da564318a468"
            body = {
                "class": "NewRelationshipRequestBody",
                "properties": {
                    "class": "SolutionLinkingWireProperties",
                    "label": "Loads",
                    "description": "Milvus is loaded from Data-Prep-Kit"
                    },
                }

            start_time = time.perf_counter()

            s_client.link_solution_linking_wire(comp1,comp2,body)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds"
                )

            assert True
        except (
                InvalidParameterException,
                PropertyServerException,
                UserNotAuthorizedException,
                ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_detach_solution_components(self):
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
                )
            s_client.create_egeria_bearer_token()
            comp1 = "b9a2735d-3a4a-4208-afe8-974e4369e733"
            comp2 = "7186fd6e-1314-47e3-bddf-da564318a468"


            start_time = time.perf_counter()

            s_client.detach_solution_linking_wire(comp1,comp2)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds"
                )

            assert True
        except (
                InvalidParameterException,
                PropertyServerException,
                UserNotAuthorizedException,
                ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_get_solution_component_by_guid(self):
        guid = '8e5aa063-468a-4504-a93f-5960f8bcee71'
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            start_time = time.perf_counter()
            response = s_client.get_solution_component_by_guid(guid, output_format='JSON')
            duration = time.perf_counter() - start_time
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
            )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
            PyegeriaException,
        ) as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_get_solution_components_by_name(self):
        name = "SolutionComponent::Hospital Processes::2"
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            start_time = time.perf_counter()
            response = s_client.get_solution_components_by_name(name, output_format='DICT')
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
            )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
            InvalidParameterException,
            PropertyServerException,
            UserNotAuthorizedException,
        ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_get_solution_component_implementations(self):
        guid = "2a5763d0-c540-4a59-8268-db7c88342269"
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
                )

            s_client.create_egeria_bearer_token()
            start_time = time.perf_counter()
            response = s_client.get_solution_component_implementations(guid, output_format='DICT')
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
                )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
                InvalidParameterException,
                PropertyServerException,
                UserNotAuthorizedException,
                ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_find_solution_components(self):
        filter = "Docling"
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            start_time = time.perf_counter()
            response = s_client.find_solution_components(filter, output_format='DICT', report_spec="Solution-Component-DrE")
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
            )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
            PyegeriaException
        ) as e:
            print_basic_exception(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_find_all_solution_components(self):
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            start_time = time.perf_counter()
            response = s_client.find_all_solution_components(output_format = "DICT")
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds, Type: {type(response)}, Element count is {len(response)}"
            )
            if isinstance(response, (list, dict)):
                print_json(data=response)

            elif type(response) is str:
                print("\n\n\t Response is: " + response)

            assert True
        except (
            InvalidParameterException,
            PropertyServerException,
            UserNotAuthorizedException,
        ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_link_subcomponent(self):
        guid1 = "40952607-f0fa-4daa-bde3-f9a89150d156"
        guid2 = "186ab31e-9660-4491-8131-940beee08863"
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()
            body = {
                  # "class": "SolutionComponentLinkRequestBody",
                  # "externalSourceGUID": "add guid here",
                  # "externalSourceName": "add qualified name here",
                  # "effectiveTime": {{isotime}},
                  # "forLineage": false,
                  # "forDuplicateProcessing": false,
                  # "properties": {
                  #   "typeName": "string",
                  #   "description": "add description here",
                  #   "effectiveFrom": {{isotime}},
                  #   "effectiveTo": {{isotime}}
                  # }
                }


            start_time = time.perf_counter()
            s_client.link_subcomponent(guid1,guid2,body)
            duration = time.perf_counter() - start_time
            print(
                f"\n\tDuration was {duration:.2f} seconds"
            )

            assert True
        except (
            InvalidParameterException,
            PropertyServerException,
            UserNotAuthorizedException,
        ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_call_execution_flow(self):
        try:
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()

            out = s_client.get_find_solution_components_call_flow("TEXT")

            print(
                out
            )

            assert True
        except (
            InvalidParameterException,
            PropertyServerException,
            UserNotAuthorizedException,
        ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()

    def test_get_component_related_elements(self):
        try:
            guid = "1a16188a-9cba-4c86-835b-e223e97d22bb"
            s_client = SolutionArchitect(
                self.view_server, self.platform_url, self.user, self.password
            )

            s_client.create_egeria_bearer_token()

            out = s_client.get_component_related_elements(guid)

            print(
                out
            )

            assert True
        except (
            InvalidParameterException,
            PropertyServerException,
            UserNotAuthorizedException,
        ) as e:
            print_exception_response(e)
            assert False, "Invalid request"

        finally:
            s_client.close_session()


#
# This file contains the common functions and definitions used in all Egeria Hands on
# Lab Notebooks. Typically these will be system-wide functions and common environment settings
#


#
# Disable certificate checking for local development use with self-signed certificate
#

import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

import os

os.environ["CURL_CA_BUNDLE"] = ""

#
# This is the location of the file system that is used in the lab.  It defaults to a "data" subfolder in
# the platform's home folder.  Here you will finc the runtime files of the platforms and the data files
# used in various labs.  The default value can be overridden using the "fileSystemRoot" environment variable.
#
fileSystemRoot = os.environ.get("fileSystemRoot", "data")
#
# This is directory where the sample data from the Egeria distribution has been unpacked in.  The default value
# describes its location in an IntelliJ workspace.  The default value can be overridden using the "egeriaDistributionRoot"
# environment variable.
#
egeriaSampleDataRoot = os.environ.get("egeriaSampleDataRoot", ".")
#
# Flag to enable debug, this cases extra information to be printed including rest calls request and response details
# Switching this flag to True produces a very large amount of output and is not recommended.
# A targeted use of this flag is recommended, set this before and reset this after the code you would like to produce debug
#
isDebug = False
disable_ssl_warnings = True
max_paging_size = 1200
#
# Definitions of the Coco Pharmaceuticals Environment
#

# These are the three main platforms used to run Egeria's OMAG Servers
corePlatformURL = os.environ.get("corePlatformURL", "https://localhost:9443")
corePlatformName = "Core Platform"

dataLakePlatformURL = os.environ.get("dataLakePlatformURL", "https://localhost:9443")
dataLakePlatformName = "Data Lake Platform"

devPlatformURL = os.environ.get("devPlatformURL", "https://localhost:9443")
devPlatformName = "Dev Platform"


# The OMAG Server Platforms host one to many OMAG Servers.  An OMAG Server could be
# a metadata server or a specialized governance server.  Its behavior is determined
# by a configuration document that defines which OMAG services are activated.
# All OMAG Server Platforms support the administration md_commands to define a server's
# configuration document.  It is also possible to create configuration documents
# through admin calls to one OMAG Server Platform and then deploy them to the
# OMAG Server Platform where they are to run.  In the Egeria hands on lab, the
# OMAG Server configuration is created on the dev platform and deployed to the
# production platforms as needed.

adminPlatformURL = devPlatformURL

# Gary Geeke is the IT Administration Lead at Coco Pharmaceuticals.
# He does all of the configuration for the OMAG Servers. Other users are introduced and make
# calls to the server as required

adminUserId = "garygeeke"
petersUserId = "peterprofile"
erinsUserId = "erinoverview"
calliesUserId = "calliequartile"
faithsUserId = "faithbroker"
harrysUserId = "harryhopeful"

# These are the names of the metadata servers used by Coco Pharmaceuticals.  Each metadata
# server runs as an OMAG Server on one of the OMAG Server Platforms

cocoMDS1PlatformURL = dataLakePlatformURL
cocoMDS1PlatformName = dataLakePlatformName
cocoMDS1Name = "cocoMDS1"

cocoMDS2PlatformURL = corePlatformURL
cocoMDS2PlatformName = corePlatformName
cocoMDS2Name = "cocoMDS2"

cocoMDS3PlatformURL = corePlatformURL
cocoMDS3PlatformName = corePlatformName
cocoMDS3Name = "cocoMDS3"

cocoMDS4PlatformURL = dataLakePlatformURL
cocoMDS4PlatformName = dataLakePlatformName
cocoMDS4Name = "cocoMDS4"

cocoMDS5PlatformURL = corePlatformURL
cocoMDS5PlatformName = corePlatformName
cocoMDS5Name = "cocoMDS5"

cocoMDS6PlatformURL = corePlatformURL
cocoMDS6PlatformName = corePlatformName
cocoMDS6Name = "cocoMDS6"

cocoMDSxPlatformURL = devPlatformURL
cocoMDSxPlatformName = devPlatformName
cocoMDSxName = "cocoMDSx"

cocoOLS1PlatformName = dataLakePlatformName
cocoOLS1PlafformURL = dataLakePlatformURL
cocoOLS1Name = "cocoOLS1"

# The open metadata servers are linked together through the following open metadata cohorts.
# The servers linked via a cohort can exchange open metadata either through federated
# queries or metadata replication.

cocoCohort = "cocoCohort"
devCohort = "devCohort"
iotCohort = "iotCohort"
ctsCohort = "ctsCohort"

# This is the view server that runs the services that support Egeria's user interface.

cocoView1PlatformURL = dataLakePlatformURL
cocoView1PlatformName = dataLakePlatformName
cocoView1Name = "cocoView1"

# These are the names of the governance servers used in Coco Pharmaceuticals' data lake.
# Each governance server runs as an OMAG Server on the Data Lake OMAG Server Platform.
# They also connect to a metadata server to retrieve their configuration and store their
# results.

governDL01PlatformURL = dataLakePlatformURL
governDL01PlatformName = dataLakePlatformName
governDL01Name = "governDL01"
governDL01ServerType = "Engine Host"
governDL01MDS = "cocoMDS1"

exchangeDL01PlatformURL = dataLakePlatformURL
exchangeDL01PlatformName = dataLakePlatformName
exchangeDL01Name = "exchangeDL01"
exchangeDL01ServerType = "Integration Daemon"
exchangeDL01MDS = "cocoMDS1"

monitorGov01PlatformURL = dataLakePlatformURL
monitorGov01PlatformName = dataLakePlatformName
monitorGov01Name = "monitorGov01"
monitorGov01ServerType = "Integration Daemon"
monitorGov01MDS = "cocoMDS1"

monitorDev01PlatformURL = devPlatformURL
monitorDev01PlatformName = devPlatformName
monitorDev01Name = "monitorDev01"
monitorDev01ServerType = "Integration Daemon"
monitorDev01MDS = "cocoMDSx"



"""
This file contains functions for extracting data from text for Egeria Markdown processing
"""
import re
import json
from typing import Any

from md_processing.md_processing_utils.common_md_utils import (print_msg, find_key_with_value, get_element_dictionary,
                                                               update_element_dictionary)
from md_processing.md_processing_utils.message_constants import INFO, EXISTS_REQUIRED
from md_processing.md_processing_utils.md_processing_constants import debug_level
from pyegeria._globals import NO_ELEMENTS_FOUND
from pyegeria.egeria_tech_client import EgeriaTech


def extract_command_plus(block: str) -> tuple[str, str, str] | None:
    """
    Extracts a multi-word object and its associated action from the given block of text.

    This function searches for a pattern in the format of `#...##` or `#...\n`
    inside the provided string `block`. The matched pattern is split into
    two parts: the action and the object type. The action is expected to
    be the first part, while the rest is treated as the object type. If
    no match is found, the function returns None.

    Lines beginning with '>' are ignored.

    Args:
        block: A string containing the block of text to search for the
            object_action and action.

    Returns:
        A tuple containing the object_action, the object type, and the object action if a
        match is found. Otherwise, returns None.
    """
    # Filter out lines beginning with '>'
    filtered_lines = [line for line in block.split('\n') if not line.strip().startswith('>')]
    filtered_block = '\n'.join(filtered_lines)

    match = re.search(r"#(.*?)(?:##|\n|$)", filtered_block)  # Using a non capturing group
    if match:
        clean_match = match.group(1).strip()
        if ' ' in clean_match:
            parts = clean_match.split(' ')
            object_action = parts[0].strip()
            # Join the rest of the parts to allow object_type to be one or two words
            object_type = ' '.join(parts[1:]).strip()
        else:
            object_type = clean_match.split(' ')[1].strip()
            object_action = clean_match.split(' ')[0].strip()

        return clean_match, object_type, object_action
    return None


def extract_command(block: str) -> str | None:
    """
    Extracts a object_action from a block of text that is contained between a single hash ('#') and
    either a double hash ('##'), a newline character, or the end of the string.

    The function searches for a specific pattern within the block of text and extracts the
    content that appears immediately after a single hash ('#'). Ensures that the extracted
    content is appropriately trimmed of leading or trailing whitespace, if present.

    Args:
        block: A string representing the block of text to process. Contains the content
            in which the object_action and delimiters are expected to be present.

    Returns:
        The extracted object_action as a string if a match is found, otherwise None.
    """
    match = re.search(r"#(.*?)(?:##|\n|$)", block)  # Using a non-capturing group
    if match:
        return match.group(1).strip()
    return None


# def extract_attribute(text: str, labels: set) -> str | None:
#     """
#         Extracts the attribute value from a string.
#
#         Args:
#             text: The input string.
#             labels: List of equivalent labels to search for
#
#         Returns:
#             The value of the attribute, or None if not found.
#
#         Note:
#             Lines beginning with '>' are ignored.
#         """
#     # Iterate over the list of labels
#     for label in labels:
#         # Construct pattern for the current label
#         # text = re.sub(r'\s+', ' ', text).strip() # just added
#         # text = re.sub(r'\n\n+', '\n\n', text).strip()
#
#         # Replace multiple spaces or tabs with a single space
#         normalized = re.sub(r'\s+', ' ', text)
#         # Collapse multiple blank lines into a single one
#         normalized = re.sub(r'\n\s*\n', '\n', normalized).strip()
#
#         # label = label.strip()
#         # pattern = rf"##\s*{re.escape(label)}\s*\n(?:\s*\n)*?(.*?)(?:#|___|$)"
#         # Normalize the label
#         normalized_label = re.sub(r'\s+', ' ', label.strip())
#
#         # Construct the regex pattern
#         pattern = rf"##\s*{re.escape(normalized_label)}\s*\n(?:\s*\n)*?(.*?)(?:#|___|$)"
#         # pattern = rf"##\s+{re.escape(label)}\n(.*?)(?:#|___|$)"  # modified from --- to enable embedded tables
#         match = re.search(pattern, text, re.DOTALL)
#         if match:
#             # Extract matched text
#             matched_text = match.group(1)
#
#             # Filter out lines beginning with '>'
#             filtered_lines = [line for line in matched_text.split('\n') if not line.strip().startswith('>')]
#             filtered_text = '\n'.join(filtered_lines)
#
#             # Replace consecutive \n with a single \n
#             extracted_text = re.sub(r'\n+', '\n', filtered_text)
#             if not extracted_text.isspace() and extracted_text:
#                 return extracted_text.strip()  # Return the cleaned text - I removed the title casing
#
#     return None


from typing import Optional, List

def extract_attribute(text: str, labels: List[str]) -> Optional[str]:
    """
    def extract_attribute(text: str, labels: List[str]) -> Optional[str]:

    Extracts the attribute value from a string while:
    - Preserving single newlines within the matched text.
    - Removing lines starting with '>'.

    Args:
        text: The input string containing labeled sections.
        labels: List of equivalent labels to search for.

    Returns:
        The cleaned value of the attribute, or None if not found.
    """
    for label in labels:
        # Construct pattern for the current label
        pattern = rf"## {re.escape(label)}\n(.*?)(?=^##|\Z)"  # Captures content until the next '##' or end of text
        match = re.search(pattern, text, re.DOTALL | re.MULTILINE)
        if match:
            # Extract matched text
            extracted_text = match.group(1)

            # Remove lines starting with '>'
            filtered_lines = [line for line in extracted_text.splitlines() if not line.lstrip().startswith(">")]

            # Join the lines back, preserving single newlines
            cleaned_text = "\n".join(filtered_lines).strip()

            if cleaned_text:
                return cleaned_text  # Return the cleaned and formatted text
    return None




def process_simple_attribute(txt: str, labels: set, if_missing: str = INFO) -> str | None:
    """Process a simple attribute based on the provided labels and if_missing value.
       Extract the attribute value from the text and return it if it exists.
       If it doesn`t exist, return None and print an error message with severity of if_missing.

       Parameters:
       ----------
       txt: str
         The block of object_action text to extract attributes from.
       labels: list
         The possible attribute labels to search for. The first label will be used in messages.
       if_missing: str, default is INFO
         Can be one of "WARNING", "ERROR", "INFO". The severity of the missing attribute.
    """
    if if_missing not in ["WARNING", "ERROR", "INFO"]:
        print_msg("ERROR", "Invalid severity for missing attribute", debug_level)
        return None

    attribute = extract_attribute(txt, labels)

    if attribute is None:
        if if_missing == INFO:
            msg = f"Optional attribute with labels `{labels}` missing"
        else:
            msg = f"Missing attribute with labels `{labels}` "
        print_msg(if_missing, msg, debug_level)
        return None
    return attribute.strip()


# def process_simple_attribute(txt: str, labels: list[str], if_missing: str = INFO) -> str | None:
#     """
#     Processes a simple attribute from a string.
#
#     Args:
#         txt: The input string.
#         labels: List of equivalent labels to search for
#         if_missing: The message level to use if the attribute is missing.
#
#     Returns:
#         The value of the attribute, or None if not found.
#     """
#     from md_processing.md_processing_utils.common_utils import debug_level, print_msg
#
#     attribute = extract_attribute(txt, labels)
#     if attribute is None and if_missing:
#         msg = f"No {labels[0]} found"
#         print_msg(if_missing, msg, debug_level)
#     return attribute


def process_name_list(egeria_client: EgeriaTech, element_type: str, txt: str, element_labels: set) -> tuple[str,
list[Any], bool | Any, bool | None | Any] | None:
    """
    Processes a list of names specified in the given text, retrieves details for each
    element based on the provided type, and generates a list of valid qualified names.

    The function reads a text block, extracts a list of element names according to the specified
    element type, looks them up using the provided Egeria client, and classifies them as valid or
    invalid. It returns the processed names, a list of qualified names, and validity and existence
    flags.

    Args:

        egeria_client (EgeriaTech): The client instance to connect and query elements from an
            external system.
        Element_type (str): The type of element, such as schema or attribute, to process.
        Txt (str): The raw input text containing element names to be processed.
        element_labels: a list of equivalent label names to use in processing the element.

    Returns:
        tuple[str | None, list | None, bool, bool]: A tuple containing:
            - Concatenated valid input names as a single string (or None if empty).
            - A list of known qualified names extracted from the processed elements.
            - A boolean indicating whether all elements are valid.
            - A boolean indicating whether all elements exist.
    """
    valid = True
    exists = True
    elements = ""
    new_element_list = []

    elements_txt = extract_attribute(txt, element_labels)

    if elements_txt is None:
        msg = f"No {element_type} found"
        print_msg("DEBUG-INFO", msg, debug_level)

    else:
        element_list = re.split(r'[,\n]+', elements_txt)

        for element in element_list:
            element_el = element.strip()

            # Get the element using the generalized function
            known_q_name, known_guid, el_valid, el_exists = get_element_by_name(egeria_client, element_type, element_el)
            # print_msg("DEBUG-INFO", status_msg, debug_level)

            if el_exists and el_valid:
                elements = f"{element_el} {elements}"  # list of the input names
                new_element_list.append(known_q_name)  # list of qualified names
            elif not el_exists:
                msg = f"No {element_type} `{element_el}` found"
                print_msg("DEBUG-INFO", msg, debug_level)
                valid = False
            valid = valid if el_valid is None else (valid and el_valid)
            exists = exists and el_exists

        if elements:
            # elements += "\n"
            msg = f"Found {element_type}: {elements}"
            print_msg("DEBUG-INFO", msg, debug_level)
        else:
            msg = " Name list contains one or more invalid qualified names."
            print_msg("DEBUG-INFO", msg, debug_level)
        return elements, new_element_list, valid, exists


# def process_name_list(egeria_client, element_type: str, txt: str, element_labels: list[str]) -> tuple[list, list,
# bool, bool]:
#     """
#     Processes a list of names from a string.
#
#     Args:
#         egeria_client: The Egeria client to use for validation.
#         element_type: The type of element to process.
#         txt: The input string.
#         element_labels: List of equivalent labels to search for
#
#     Returns:
#         A tuple containing:
#         - A list of element names
#         - A list of element qualified names
#         - A boolean indicating if all elements are valid
#         - A boolean indicating if any elements exist
#     """
#     from md_processing.md_processing_utils.common_utils import debug_level, print_msg
#
#     element_names = []
#     element_q_names = []
#     all_valid = True
#     any_exist = False
#
#     # Get the list of element names
#     element_list = process_simple_attribute(txt, element_labels)
#     if element_list:
#         # Split the list by commas or newlines
#         element_names = list(filter(None, re.split(r'[,\n]+', element_list.strip())))
#
#         # Validate each element
#         for element_name in element_names:
#             element_name = element_name.strip()
#             if element_name:
#                 element = get_element_by_name(egeria_client, element_type, element_name)
#                 if element:
#                     any_exist = True
#                     element_q_name = element.get('qualifiedName', None)
#                     if element_q_name:
#                         element_q_names.append(element_q_name)
#                     else:
#                         all_valid = False
#                         msg = f"Element {element_name} has no qualified name"
#                         print_msg("ERROR", msg, debug_level)
#                 else:
#                     all_valid = False
#                     msg = f"Element {element_name} not found"
#                     print_msg("ERROR", msg, debug_level)
#
#     return element_names, element_q_names, all_valid, any_exist

def process_element_identifiers(egeria_client: EgeriaTech, element_type: str, element_labels: set, txt: str,
                                action: str, version: str = None) -> tuple[str, str, bool, bool]:
    """
    Processes element identifiers by extracting display name and qualified name from the input text,
    checking if the element exists in Egeria and validating the information.

    Parameters
    ----------
    egeria_client: EgeriaTech
        Client object for interacting with Egeria.
    element_type: str
        type of element to process (e.g., 'blueprint', 'category', 'term')
    element_labels: a list of equivalent label names to use in processing the element.
    txt: str
        A string representing the input text to be processed for extracting element identifiers.
    action: str
        The action object_action to be executed (e.g., 'Create', 'Update', 'Display', ...)
    version: str, optional = None
        An optional version identifier used if we need to construct the qualified name

    Returns: tuple[str, str, str, bool, bool]
        A tuple containing:
        - qualified_name: Empty string or element identifier
        - guid: Empty string or additional element information
        - Valid: Boolean indicating if the element information is valid
        - Exists: Boolean indicating if the element exists in Egeria
    """
    valid = True

    element_name = extract_attribute(txt, element_labels)
    qualified_name = extract_attribute(txt, ["Qualified Name"])

    if qualified_name:
        q_name, guid, unique, exists = get_element_by_name(egeria_client, element_type,
                                                           qualified_name)  # Qualified name could be different if it
        # is being updated
    else:
        q_name, guid, unique, exists = get_element_by_name(egeria_client, element_type, element_name)
    if unique is False:
        msg = f"Multiple elements named  {element_name} found"
        print_msg("DEBUG-ERROR", msg, debug_level)
        valid = False

    if action == "Update" and not exists:
        msg = f"Element {element_name} does not exist"
        print_msg("DEBUG-ERROR", msg, debug_level)
        valid = False

    elif action == "Update" and exists:
        msg = f"Element {element_name} exists"
        print_msg("DEBUG-INFO", msg, debug_level)

    elif action == "Create" and exists:
        msg = f"Element {element_name} already exists"
        print_msg("DEBUG-ERROR", msg, debug_level)
        valid = False

    elif action == "Create" and not exists:
        msg = f"{element_type} `{element_name}` does not exist"
        print_msg("DEBUG-INFO", msg, debug_level)

        if q_name is None and qualified_name is None:
            q_name = egeria_client.__create_qualified_name__(element_type, element_name, version_identifier=version)
            update_element_dictionary(q_name, {'display_name': element_name})
        elif qualified_name:
            update_element_dictionary(qualified_name, {'display_name': element_name})
    elif action == EXISTS_REQUIRED:
        if not exists:
            msg = f"Required {element_type} `{element_name}` does not exist"
            print_msg("DEBUG-ERROR", msg, debug_level)
            valid = False
        else:
            msg = f"Required {element_type} `{element_name}` exists"
            print_msg("DEBUG-INFO", msg, debug_level)
            valid = True

    return q_name, guid, valid, exists


def get_element_by_name(egeria_client, element_type: str, element_name: str) -> tuple[
    str | None, str | None, bool | None, bool | None]:
    """
    Generalized function to retrieve an element by name based on its type.

    Parameters:
    egeria_client: Client
        Client object for interacting with Egeria.
    element_type: str
        The type of element to retrieve (e.g., 'blueprint', 'category', 'term').
    element_name: str
        The name of the element to retrieve.

    Returns:
        tuple of qualified_name, guid, uniqye, exists
    """
    unique = None

    element_dict = get_element_dictionary()
    q_name = find_key_with_value(element_name)
    if q_name:  # use information from element_dictionary
        guid = element_dict[q_name].get('guid', None)
        unique = True

        if guid is not None:  # Found complete entry in element_dictionary
            msg = f'Found {element_type} qualified name and guid in element_dictionary for `{element_name}`'
            print_msg("DEBUG-INFO", msg, debug_level)
            exists = True
            return q_name, guid, unique, exists

        else:  # Missing guid from element_dictionary
            guid = egeria_client.__get_guid__(qualified_name=q_name)
            if guid == NO_ELEMENTS_FOUND:
                guid = None
                exists = False
                msg = f"No {element_type} guid found with name {element_name} in Egeria"
                print_msg("DEBUG-INFO", msg, debug_level)

                return q_name, guid, unique, exists
            else:
                exists = True
                update_element_dictionary(q_name, {'guid': guid})
                msg = f"Found guid value of {guid} for  {element_name} in Egeria"
                print_msg("DEBUG-INFO", msg, debug_level)

                return q_name, guid, unique, exists

    # Haven't seen this element before
    property_names = ['qualifiedName', 'displayName', 'title']
    open_metadata_type_name = None
    details = egeria_client.get_elements_by_property_value(element_name, property_names, open_metadata_type_name)
    if isinstance(details, str):
        msg = f"{element_type} `{element_name}` not found in Egeria"
        print_msg("DEBUG-INFO", msg, debug_level)
        exists = False
        return None, None, unique, exists
    if len(details) > 1:
        if q_name is None:
            q_name = egeria_client.__create_qualified_name__(element_type, element_name)
            guid = egeria_client.__get_guid__(qualified_name=q_name)
            update_element_dictionary(q_name, {'guid': guid})
            exists = True if guid != "No elements found" else False
            return q_name, guid, unique, exists
        else:
            msg = (f"More than one element with name {element_name} found, please specify a "
                   f"**Qualified Name**")
            print_msg("DEBUG-ERROR", msg, debug_level)
            unique = False
            exists = None
            return element_name, None, unique, exists

    el_qname = details[0]["properties"].get('qualifiedName', None)
    el_guid = details[0]['elementHeader']['guid']
    el_display_name = details[0]["properties"].get('displayName', None)
    if el_display_name is None:
        el_display_name = details[0]["properties"].get('name', None)
    update_element_dictionary(el_qname, {
        'guid': el_guid, 'displayName': el_display_name
        })
    msg = f"Found {element_type} `{el_display_name}` with qualified name `{el_qname}`"
    print_msg("DEBUG-INFO", msg, debug_level)
    exists = True
    unique = True
    return el_qname, el_guid, unique, exists


def update_a_command(txt: str, object_action: str, obj_type: str, q_name: str, u_guid: str) -> str:
    """
    Updates a command in a string.

    Args:
        txt: The input string.
        object_action: The command to update.
        obj_type: The type of object to update.
        q_name: The qualified name of the object.
        u_guid: The GUID of the object.

    Returns:
        The updated string.
    """

    # Determine the new action
    new_action = "Update" if object_action == "Create" else "Create"

    # Replace the object_action
    new_command = f"{new_action} {obj_type}"
    pattern = rf"#\s*{object_action}\s+{obj_type}"
    replacement = f"# {new_command}"
    updated_txt = re.sub(pattern, replacement, txt)

    # Add qualified name and GUID if updating
    if new_action == "Update" and q_name and u_guid:
        # Check if Qualified Name section exists
        if "## Qualified Name" not in updated_txt:
            # Add Qualified Name section before the first ## that's not part of the object_action
            pattern = r"(##\s+[^#\n]+)"
            replacement = f"## Qualified Name\n{q_name}\n\n\\1"
            updated_txt = re.sub(pattern, replacement, updated_txt, count=1)

        # Check if GUID section exists
        if "## GUID" not in updated_txt and "## guid" not in updated_txt:
            # Add GUID section before the first ## that's not part of the object_action or Qualified Name
            pattern = r"(##\s+(?!Qualified Name)[^#\n]+)"
            replacement = f"## GUID\n{u_guid}\n\n\\1"
            updated_txt = re.sub(pattern, replacement, updated_txt, count=1)

    return updated_txt
